from abc import ABC, abstractmethod

from pysketcher._point import Point
from pysketcher._style import Style


class Stylable(ABC):
    """Abstract Superclass for objects which can accept a style."""

    _style: Style

    @abstractmethod
    def __init__(self):
        self._style = Style()

    @property
    def style(self):
        """The style.

        Returns:
            The style associated with the class.
        """
        return self._style

    @style.setter
    def style(self, style: Style):
        self._style = style

    def set_line_width(self, line_width: float) -> "Stylable":
        """Set the line width.

        Args:
            line_width: The width of the line.

        Returns:
            The object with its style modified with the new value of ``line_width``.
        """
        self.style.line_width = line_width
        return self

    def set_line_style(self, line_style: Style.LineStyle) -> "Stylable":
        """Set the line style.

        Args:
            line_style: The width of the line.

        Returns:
            The object with its style modified with the new value of ``line_style``.
        """
        self.style.line_style = line_style
        return self

    def set_line_color(self, line_color: Style.Color) -> "Stylable":
        """Set the line color.

        Args:
            line_color: The color of the line.

        Returns:
            The object with its style modified with the new value of ``line_color``.
        """
        self.style.line_color = line_color
        return self

    def set_fill_pattern(self, fill_pattern: Style.FillPattern) -> "Stylable":
        """Set the fill_pattern style.

        Args:
            fill_pattern: The width of the line.

        Returns:
            The object with its style modified with the new value of ``line_style``.
        """
        self.style.fill_pattern = fill_pattern
        return self

    def set_fill_color(self, fill_color: Style.Color) -> "Stylable":
        """Set the fill color.

        Args:
            fill_color: The width of the line.

        Returns:
            The object with its style modified with the new value of ``fill_color``.
        """
        self.style.fill_color = fill_color
        return self

    def set_arrow(self, arrow: Style.ArrowStyle) -> "Stylable":
        """Set the arrows which should adorn the object.

        Args:
            arrow: The nature of the arrow.

        Returns:
            The object with its style modified with the new value of ``arrow``.
        """
        self.style.arrow = arrow
        return self

    def set_shadow(self, shadow: float) -> "Stylable":
        """Set the shadow.

        Args:
            shadow: The distance of the shadow from the object.

        Returns:
            The object with its style modified with the new value of ``shadow``.
        """
        self.style.shadow = shadow
        return self


class Shape(Stylable):
    """Superclass for drawing different geometric shapes.

    Subclasses define shapes, but drawing, rotation, translation,
    etc. are done in generic functions in this superclass.
    """

    @abstractmethod
    def __init__(self):
        super().__init__()

    @abstractmethod
    def rotate(self, angle: float, center: Point) -> "Shape":
        """Rotate the shape.

        Args:
            angle: The ``Angle`` in radians through which the shape should be rotated.
            center: The ``Point`` about which the rotation should be performed.

        Raises:
            NotImplementedError: when shape does not implement ``rotate``.
        """
        raise NotImplementedError

    def translate(self, vec) -> "Shape":
        """Translate the shape.

        Args:
            vec: The vector through which the ``Shape`` should be translated.

        Raises:
            NotImplementedError: when shape does not implement ``translate``.
        """
        raise NotImplementedError

    def scale(self, factor) -> "Shape":
        """Scale the shape.

        Args:
            factor: The factor by which the ``Shape`` should be scaled.

        Raises:
            NotImplementedError: when shape does not implement ``scale``.
        """
        raise NotImplementedError



"""Provides a means to group primitive shapes.

Primitives can be combined together to form reusable
groups which can then be transformed. This module also
provides a set of pre-baked primitives for convenience.
"""

from pysketcher.composition._composition import Composition

__all__ = ["Composition"]



"""Provides a means to annotate shapes.

Annotations are text objects which are positioned using the dimensions
of a provide shape.
"""

from pysketcher.annotation._arc_annotation import ArcAnnotation
from pysketcher.annotation._line_annotation import LineAnnotation
from pysketcher.annotation._text_position import TextPosition

__all__ = ["ArcAnnotation", "LineAnnotation", "TextPosition"]



from enum import auto, Enum, unique

import numpy as np

from pysketcher._angle import Angle
from pysketcher._arc import Arc
from pysketcher._style import Style
from pysketcher.annotation import ArcAnnotation
from pysketcher.composition import Composition


class Moment(Composition):
    r"""A symbol which represents a moment.

    This is an ``ArcWithText`` with the ``arc_angle`` fixed at :math:`\pi`.

    Args:
        text: The text to display.
        center: The centre of the moment.
        radius: The radius of the moment.
        start_angle: The angle from the +ve horizontal at which the moment should
            start.
        text_spacing: The spacing of the text.
        resolution: The number of points on the arc.

    Examples:
        >>> moment = ps.Moment("$M$", ps.Point(0, 0), 1.0)
        >>> fig = ps.Figure(-1.2, 1.2, -1.2, 1.2, backend=MatplotlibBackend)
        >>> fig.add(moment)
        >>> fig.save("pysketcher/images/moment.png")

        .. figure:: images/moment.png
            :alt: An example of Moment.
            :figclass: align-center

            An example of ``Moment``.
    """

    @unique
    class Direction(Enum):
        """Indicates the direction of the moment."""

        CLOCKWISE = auto()
        COUNTER_CLOCKWISE = auto()

    @unique
    class Orientation(Enum):
        """Indicated the orientation of the moment."""

        LEFT = auto()
        RIGHT = auto()

    _DEFAULT_DIRECTION: Direction = Direction.COUNTER_CLOCKWISE
    _DEFAULT_ORIENTATION: Orientation = Orientation.LEFT

    def __init__(
        self,
        text,
        center,
        radius,
    ):
        self._direction = self._DEFAULT_DIRECTION
        self._orientation = self._DEFAULT_ORIENTATION

        style = (
            Style.ArrowStyle.END
            if self._direction == self.Direction.COUNTER_CLOCKWISE
            else Style.ArrowStyle.START
        )

        start_angle = (
            np.pi / 2 if self._orientation == self.Orientation.LEFT else -np.pi / 2
        )

        self._arc = Arc(center, radius, start_angle, Angle(np.pi)).set_arrow(style)
        self._label = ArcAnnotation(text, self._arc)
        super().__init__({"arc": self._arc, "label": self._label})



import numpy as np

from pysketcher._curve import Curve
from pysketcher._point import Point
from pysketcher._spline import Spline
from pysketcher._style import Style


def _scale_array(min: float, max: float, ps: np.array):
    return min - ps.min() + ps * (max - min) / (ps.max() - ps.min())


class SketchyFunc1(Spline):
    """A typical function curve used to illustrate an "arbitrary" function.

    Examples:
        >>> f = ps.SketchyFunc1()
        >>> fig = ps.Figure(0.0, 7.0, 0.0, 3.0, backend=MatplotlibBackend)
        >>> fig.add(f)
        >>> fig.save("pysketcher/images/sketchyfunc1.png")

        .. figure:: images/sketchyfunc1.png
            :alt: An example of an SketchyFunc1.
            :figclass: align-center

            An example of ``SketchyFunc1``.
    """

    domain = [1, 6]

    def __init__(self, name=None, name_pos="start", x_min=0, x_max=6, y_min=0, y_max=2):
        xs = np.array([0.0, 2.0, 3.0, 4.0, 5.0, 6.0])
        ys = np.array([1, 1.8, 1.2, 0.7, 0.8, 0.85])

        xs = _scale_array(x_min, x_max, xs)
        ys = _scale_array(y_min, y_max, ys)
        points = Point.from_coordinate_lists(xs, ys)

        super().__init__(points)


class SketchyFunc2(Curve):
    """A typical function curve used to illustrate an "arbitrary" function.

    Examples:
        >>> f = ps.SketchyFunc2()
        >>> fig = ps.Figure(0.0, 3.0, 0.0, 1.5, backend=MatplotlibBackend)
        >>> fig.add(f)
        >>> fig.save("pysketcher/images/sketchyfunc2.png")

        .. figure:: images/sketchyfunc2.png
            :alt: An example of an SketchyFunc1.
            :figclass: align-center

            An example of ``SketchyFunc2``.
    """

    domain = [0, 2.25]

    def __init__(self, x_min=0, x_max=2.25, y_min=0.046679703125, y_max=1.259375):
        a = 0
        b = 2.25
        resolution = 100
        xs = np.linspace(a, b, resolution + 1)
        ys = self.__call__(xs)
        # Scale x and y
        xs = _scale_array(x_min, x_max, xs)
        ys = _scale_array(y_min, y_max, ys)
        points = Point.from_coordinate_lists(xs, ys)

        super().__init__(points)

    def __call__(self, x):
        """Returns the value of the function at a given :math:`x`."""
        return 0.5 + x * (2 - x) * (0.9 - x)  # on [0, 2.25]


class SketchyFunc3(Spline):
    """A typical function curve used to illustrate an "arbitrary" function.

    Examples:
        >>> f = ps.SketchyFunc3()
        >>> fig = ps.Figure(0.0, 7.0, 0.0, 4.5, backend=MatplotlibBackend)
        >>> fig.add(f)
        >>> fig.save("pysketcher/images/sketchyfunc3.png")

        .. figure:: images/sketchyfunc3.png
            :alt: An example of an SketchyFunc3.
            :figclass: align-center

            An example of ``SketchyFunc3``.
    """

    domain = [0, 6]

    def __init__(self, x_min=0, x_max=6, y_min=0.5, y_max=3.8):
        xs = np.array([0.0, 2.0, 3.0, 4.0, 5.0, 6.0])
        ys = np.array([0.5, 3.5, 3.8, 2, 2.5, 3.5])

        # Scale x and y
        xs = _scale_array(x_min, x_max, xs)
        ys = _scale_array(y_min, y_max, ys)
        points = Point.from_coordinate_lists(xs, ys)

        super().__init__(points)
        self.style.line_color = Style.Color.BLACK


class SketchyFunc4(Spline):
    """A typical function curve used to illustrate an "arbitrary" function.

    Can be a companion function to SketchyFunc3.

    Examples:
        >>> f = ps.SketchyFunc4()
        >>> fig = ps.Figure(0.0, 7.0, 0.0, 3.0, backend=MatplotlibBackend)
        >>> fig.add(f)
        >>> fig.save("pysketcher/images/sketchyfunc4.png")

        .. figure:: images/sketchyfunc4.png
            :alt: An example of an SketchyFunc4.
            :figclass: align-center

            An example of ``SketchyFunc4``.
    """

    domain = [1, 6]

    def __init__(self, name_pos="start", x_min=0, x_max=6, y_min=0.5, y_max=1.8):
        xs = np.array([0, 2, 3, 4, 5, 6])
        ys = np.array([1.5, 1.3, 0.7, 0.5, 0.6, 0.8])
        # Scale x and y
        xs = _scale_array(x_min, x_max, xs)
        ys = _scale_array(y_min, y_max, ys)
        points = Point.from_coordinate_lists(xs, ys)

        super().__init__(points)



import numpy as np

from pysketcher._circle import Circle
from pysketcher._line import Line
from pysketcher._point import Point
from pysketcher.composition import Composition


class Wheel(Composition):
    """A representation of a wheel.

    Args:
        center: The center of the ``Wheel``.
        radius: The radius of the ``Wheel``.
        inner_radius: The radius of the hub of the ``Wheel``.

    Examples:
        >>> w = ps.Wheel(ps.Point(1.0, 1.0), 0.5, 0.25)
        >>> fig = ps.Figure(0.0, 2.0, 0.0, 2.0, backend=MatplotlibBackend)
        >>> fig.add(w)
        >>> fig.save("pysketcher/images/wheel.png")

        .. figure:: images/wheel.png
            :alt: An example of a Wheel.
            :figclass: align-center

            An example of a ``Wheel``.
    """

    _center: Point
    _radius: float
    _inner_radius: float
    _nlines: int

    def __init__(
        self, center: Point, radius: float, inner_radius: float = None, nlines: int = 10
    ):
        self._center = center
        self._radius = radius
        self._inner_radius = inner_radius
        self._nlines = nlines

        if inner_radius is None:
            self._inner_radius = radius / 5.0

        outer = Circle(center, radius)
        inner = Circle(center, inner_radius)
        lines = []
        # Draw nlines+1 since the first and last coincide
        # (then nlines lines will be visible)
        t = np.linspace(0, 2 * np.pi, nlines + 1)

        xinner = self._center.x + self._inner_radius * np.cos(t)
        yinner = self._center.y + self._inner_radius * np.sin(t)
        xouter = self._center.x + self._radius * np.cos(t)
        youter = self._center.y + self._radius * np.sin(t)
        lines = [
            Line(Point(xi, yi), Point(xo, yo))
            for xi, yi, xo, yo in zip(xinner, yinner, xouter, youter)
        ]
        super().__init__(
            {
                "inner": inner,
                "outer": outer,
                "spokes": Composition(
                    {"spoke%d" % i: lines[i] for i in range(len(lines))}
                ),
            }
        )



from abc import ABC, abstractmethod
from typing import Callable, Tuple, Union

from pysketcher._drawable import Drawable


class Backend(ABC):
    pass

    @abstractmethod
    def add(self, shape: Drawable) -> None:
        pass

    @abstractmethod
    def erase(self) -> None:
        pass

    @abstractmethod
    def show(self) -> None:
        pass

    @abstractmethod
    def save(self, filename: str) -> None:
        pass

    def animate(
        self,
        func: Callable[[float], Drawable],
        interval: Union[Tuple[float, float], Tuple[float, float, float]],
    ):
        raise NotImplementedError("This backend doesn't implement animation.")

    def save_animation(self, filename: str):
        raise NotImplementedError("This backend doesn't implement animation.")



from typing import List, Tuple

import numpy as np

from pysketcher._angle import Angle


class Point:
    """Immutable Point class which implements basic point arithmetic.

    Args:
        x: float
            The x co-ordinate
        y: float
            The y co-ordinate
    """

    _x: float
    _y: float

    def __init__(self, x: float, y: float):
        self._x = float(x)
        self._y = float(y)

    def __add__(self, other: "Point") -> "Point":
        """Sums the co-ordinates of the two points.

        Args:
            other: the point to add to ``self``.

        Returns:
            The sum of ``self`` and ``other``.
        """
        return Point(self.x + other._x, self.y + other.y)

    def __sub__(self, other: "Point") -> "Point":
        """Subtracts the co-ordinates of the two points.

        Args:
            other: the point to subtract from ``self``.

        Returns:
            The difference between ``self`` and ``other``.
        """
        return Point(self.x - other._x, self.y - other.y)

    def __mul__(self, scalar: float) -> "Point":
        """Multiplies ``self`` by ``scalar``.

        Args:
            scalar: the factor to scale ``self`` by.

        Returns:
            A point with each co-ordinate scaled from ``self`` by ``scalar``.
        """
        return Point(self.x * scalar, self.y * scalar)

    def __abs__(self) -> float:
        """The pythagorean length of ``self``.

        Returns:
            The cartesian distance from ``Point(0,0)`` to ``self``.
        """
        return np.hypot(self.x, self.y)

    def __eq__(self, other: "Point") -> bool:
        """Implementation of point equality.

        Note that owing to the vagaries of floating point arithmetic
        two points which have co-ordinates within 1e-4 will be
        considered equal.

        Args:
            other: the ``Point`` to compare to ``self``.

        Returns:
            True if ``Point`` is equal to ``self``.
        """
        return self._is_close(self.x, other.x) and self._is_close(self.y, other.y)

    def __str__(self):
        """A string representation of ``self``."""
        return self.__repr__()

    def __repr__(self):
        """An unambiguous string representation of ``self``."""
        return "P(%s, %s)" % (
            np.format_float_scientific(self.x),
            np.format_float_scientific(self.y),
        )

    @staticmethod
    def _is_close(x: float, y: float) -> bool:
        return np.isclose(x, y, atol=1e-4)

    @property
    def x(self) -> float:
        """The x co-ordinate of the Point."""
        return self._x

    @property
    def y(self) -> float:
        """The y co-ordinate of the Point."""
        return self._y

    @property
    def unit_vector(self) -> "Point":
        """Returns a ``Point`` of length 1 in the direction of the ``Point``."""
        if self._is_close(abs(self), 0.0):
            raise ZeroDivisionError("Length of Vector cannot be Zero")
        return self * (1 / (abs(self)))

    @property
    def angle(self) -> Angle:
        """Returns the ``Angle`` the ``Point`` makes to the +ve horizontal."""
        angle = Angle(np.arctan2(self.y, self.x))
        return angle

    @property
    def radius(self) -> float:
        """Returns the distance from ``Point(0.,0.)`` to this point."""
        return abs(self)

    @property
    def normal(self) -> "Point":
        r"""Returns a ``Point`` of length 1, :math:``$\frac{\pi}{2}$`` from ``self``."""
        uv = self.unit_vector
        return Point(-uv.y, uv.x)

    def rotate(self, angle: Angle, center: "Point") -> "Point":
        """Returns a ``Point``, ``angle`` (in radians) around (`x`,`y`).

        Args:
            angle: The angle through which the rotation should be made.
            center: The point about which the rotation should be made.

        Returns:
            The rotated point.
        """
        if not type(angle) == Angle:
            angle = Angle(angle)

        # Check for a few degenerate cases:
        if angle == Angle(0.0):
            p = self
        elif angle == Angle(np.pi / 2):
            p = Point(center.x - self.y + center.y, center.y + self.x - center.x)
        elif angle == Angle(np.pi):
            p = Point(2 * center.x - self.x, 2 * center.y - self.y)
        elif angle == Angle(-np.pi / 2):
            p = Point(center.x + self.y - center.y, center.y - self.x + center.x)
        else:
            c = np.cos(angle)
            s = np.sin(angle)
            p = Point(
                center.x + (self.x - center.x) * c - (self.y - center.y) * s,
                center.y + (self.x - center.x) * s + (self.y - center.y) * c,
            )
        return p

    def scale(self, factor: float) -> "Point":
        """Scale point coordinates by `factor`: ``x = factor*x``, etc.

        Args:
            factor: the amount the ``Point`` should be scaled by.

        Returns:
            The scaled point.
        """
        return self * factor

    def translate(self, vec: "Point") -> "Point":
        """Translate point by a vector `vec`.

        Args:
            vec: The vector (``Point``) through which the point should be
                translated.

        Returns:
            The translated vector.
        """
        return self + vec

    @staticmethod
    def from_coordinate_lists(xs: List[float], ys: List[float]) -> List["Point"]:
        """Generates points from lists of co-ordinates.

        Args:
            xs: the x co-ordinates
            ys: the y co-ordinates

        Returns:
            A list of ``Point`` made up of the respective co-ordinates.

        Raises:
            ValueError: When the co-ordinates lists are of different lengths.
        """
        if len(xs) != len(ys):
            raise ValueError("xs and ys must be the same length")
        return [Point(xs[i], ys[i]) for i in range(len(xs))]

    @staticmethod
    def to_coordinate_lists(ps: List["Point"]) -> Tuple[List[float], List[float]]:
        """Generates lists of co-ordinates from points."""
        xs = [p.x for p in ps]
        ys = [p.y for p in ps]
        return xs, ys



from copy import copy
from typing import List

import matplotlib.pyplot as plt
import numpy as np

import pysketcher as ps
from pysketcher._style import Style
from pysketcher.backend.matplotlib._matplotlib_adapter import MatplotlibAdapter
from pysketcher.backend.matplotlib._matplotlib_style import MatplotlibStyle


class MatplotlibCurve(MatplotlibAdapter):
    types: List[type] = [ps.Curve]

    def plot(self, curve: ps.Curve, axes: plt.Axes) -> None:
        """Draw a curve with coordinates x and y (arrays)."""
        mpl_style = MatplotlibStyle(curve.style)

        xs = curve.xs
        ys = curve.ys

        if len(xs) == 2 and curve.style.arrow == Style.ArrowStyle.DOUBLE:
            # For multi-segment curves we can just make the first segment an
            # arrow if the style is START or DOUBLE and the last segment an
            # arrow if the style is END or DOUBLE. If there is only one segment
            # and the style is DOUBLE this won't work, so we split into two segments
            xs = np.linspace(xs[0], xs[1], 3)
            ys = np.linspace(ys[0], ys[1], 3)

        if curve.style.arrow is not None:
            if curve.style.arrow in [Style.ArrowStyle.START, Style.ArrowStyle.DOUBLE]:
                x_s, y_s = xs[0], ys[0]
                dx_s, dy_s = xs[1] - xs[0], ys[1] - ys[0]
                start_style = copy(curve.style)
                start_style.arrow = Style.ArrowStyle.END
                self._plot_arrow(x_s, y_s, dx_s, dy_s, start_style, axes)
                xs = xs[1:]
                ys = ys[1:]
            if curve.style.arrow in [Style.ArrowStyle.END, Style.ArrowStyle.DOUBLE]:
                x_e, y_e = xs[-2], ys[-2]
                dx_e, dy_e = xs[-1] - xs[-2], ys[-1] - ys[-2]
                end_style = copy(curve.style)
                end_style.arrow = Style.ArrowStyle.START
                self._plot_arrow(x_e, y_e, dx_e, dy_e, end_style, axes)
                xs = xs[:-1]
                ys = ys[:-1]

        if len(xs) >= 2:
            [line] = axes.plot(
                xs,
                ys,
                mpl_style.line_color,
                linewidth=mpl_style.line_width,
                linestyle=mpl_style.line_style,
            )
            if mpl_style.fill_color or mpl_style.fill_pattern:
                [line] = plt.fill(
                    xs,
                    ys,
                    mpl_style.fill_color,
                    edgecolor=mpl_style.line_color,
                    linewidth=mpl_style.line_width,
                    hatch=mpl_style.fill_pattern,
                )

        # if mpl_style.shadow:
        # http://matplotlib.sourceforge.net/users/transforms_tutorial.html
        # #using-offset-transforms-to-create-a-shadow-effect
        # shift the object over 2 points, and down 2 points
        #     dx, dy = mpl_style.shadow / 72.0, -mpl_style.shadow / 72.0
        #     offset = transforms.ScaledTranslation(dx, dy, fig.dpi_scale_trans)
        #     shadow_transform = axes.transData + offset
        #     # now plot the same data with our offset transform;
        #     # use the zorder to make sure we are below the line
        #     axes.plot(
        #         x,
        #         y,
        #         linewidth=mpl_style.line_width,
        #         color="gray",
        #         transform=shadow_transform,
        #         zorder=0.5 * line.get_zorder(),
        #     )

    def _plot_arrow(self, x, y, dx, dy, style: Style, axes: plt.Axes):
        """Draw arrow (dx,dy) at (x,y). `style` is '->', '<-' or '<->'."""
        if style.arrow == Style.ArrowStyle.DOUBLE:
            raise ValueError("Only a single ended arrow is supported by this method.")
        mpl_style = MatplotlibStyle(style)
        if style.arrow == Style.ArrowStyle.END:
            x = x + dx
            y = y + dy
            dx = -dx
            dy = -dy
        axes.arrow(
            x,
            y,
            dx,
            dy,
            facecolor=mpl_style.line_color,
            edgecolor=mpl_style.line_color,
            linestyle=mpl_style.line_style,
            linewidth=mpl_style.line_width,
            head_width=0.05,
            # head_width=self.arrow_head_width,
            # width=1,  # width of arrow body in coordinate scale
            length_includes_head=True,
            shape="full",
        )



from enum import auto, Enum, unique


@unique
class Color(Enum):
    """A set of colors supported across all backends."""

    RED = auto()
    GREEN = auto()
    BLUE = auto()
    CYAN = auto()
    MAGENTA = auto()
    YELLOW = auto()
    WHITE = auto()
    BLACK = auto()



from pysketcher._arrow import Arrow
from pysketcher._point import Point
from pysketcher._rectangle import Rectangle
from pysketcher.composition import Composition


class UniformLoad(Composition):
    """Downward-pointing arrows indicating a vertical load.

    The arrows are of equal length and filling a rectangle
    specified as in the :class:`Rectangle` class.

    Examples:
        >>> l = ps.UniformLoad(ps.Point(0.5, 0.5), 4, 0.5)
        >>> fig = ps.Figure(0.0, 5.0, 0.0, 1, backend=MatplotlibBackend)
        >>> fig.add(l)
        >>> fig.save("pysketcher/images/uniform_load.png")

        .. figure:: images/uniform_load.png
            :alt: An example of a Uniform Load.
            :figclass: align-center

            An example of a ``UniformLoad``.
    """

    def __init__(
        self, lower_left_corner: Point, width: float, height: float, num_arrows=10
    ):
        self._lower_left_corner = lower_left_corner
        self._width = width
        self._height = height
        box = Rectangle(lower_left_corner, width, height)
        shapes = {"box": box}
        dx = float(width) / (num_arrows - 1)
        for i in range(num_arrows):
            x = lower_left_corner.x + i * dx
            start = Point(x, lower_left_corner.y + height)
            end = Point(x, lower_left_corner.y)
            shapes["arrow%d" % i] = Arrow(start, end)
        super().__init__(shapes)

    @property
    def mid_top(self) -> Point:
        """The middle of the top of the load."""
        return self._lower_left_corner + Point(self._width / 2, self._height)



from pysketcher._curve import Curve
from pysketcher._line import Line
from pysketcher._point import Point
from pysketcher._rectangle import Rectangle
from pysketcher._style import Style
from pysketcher.composition import Composition


class Dashpot(Composition):
    """A representation of a dashpot.

    Specify a vertical dashpot of height `total_length` and `start` as
    bottom/starting point. The first bar part has length `bar_length`.
    Then comes the dashpot as a rectangular construction of total
    width `width` and height `dashpot_length`. The position of the
    piston inside the rectangular dashpot area is given by
    `piston_pos`, which is the distance between the first bar (given
    by `bar_length`) to the piston.
    If some of `dashpot_length`, `bar_length`, `width` or `piston_pos`
    are not given, suitable default values are calculated.

    Args:
        start: The starting point of the Dashpot.
        total_length: The total length of the Dashpot including bar.
        bar_length: The length of the bar.
        width: The width of the dashpot.
        dashpot_length: The length of the Dashpot excluding bar.
        piston_pos: The position of the piston. Defaults to 1/3 length.

    Examples:
        >>> L = 12.0
        >>> H = L / 6.0
        >>> x = L / 2.0
        >>> d_start = ps.Point(0, 0)
        >>> d = ps.Dashpot(d_start, L + x)
        >>> fig = ps.Figure(-3, 3, -1, 13, backend=MatplotlibBackend)
        >>> fig.add(d)
        >>> fig.save("pysketcher/images/dashpot.png")

        .. figure:: images/dashpot.png
            :alt: An example of Dashpot.
            :figclass: align-center
            :scale: 30%

            An example of ``Dashpot``.
    """

    _dashpot_fraction = 1.0 / 2  # fraction of total_length
    _piston_gap_fraction = 1.0 / 6  # fraction of width
    _piston_thickness_fraction = 1.0 / 8  # fraction of dashplot_length

    def __init__(
        self,
        start: Point,
        total_length: float,
        bar_length: float = None,
        width: float = None,
        dashpot_length: float = None,
        piston_pos: float = None,
    ):
        B = start
        L = total_length
        if width is None:
            w = L / 10.0  # total width 1/5 of length
        else:
            w = width / 2.0
        s = bar_length

        shapes = {}
        # dashpot is p0-p1 in y and width 2*w
        if dashpot_length is None:
            if s is None:
                f = Dashpot._dashpot_fraction
                s = L * (1 - f) / 2.0  # default
            p1 = Point(B.x, B.y + L - s)
            dashpot_length = f * L
        else:
            if s is None:
                f = 1.0 / 2  # the bar lengths are taken as f*dashpot_length
                s = f * dashpot_length  # default
            p1 = Point(B.x, B.y + s + dashpot_length)
        p0 = Point(B.x, B.y + s)
        p2 = Point(B.x, B.y + L)

        if not (p2.y > p1.y > p0.y):
            raise ValueError(
                (
                    "Dashpot has inconsistent dimensions! start: %g, "
                    "dashpot begin: %g, dashpot end: %g, very end: %g"
                    % (B.y, p0.y, p1.y, p2.y)
                )
            )

        shapes["line start"] = Line(B, p0)

        shapes["pot"] = Curve(
            [
                Point(p1.x - w, p1.y),
                Point(p0.x - w, p0.y),
                Point(p0.x + w, p0.y),
                Point(p1.x + w, p1.y),
            ]
        )
        piston_thickness = dashpot_length * Dashpot._piston_thickness_fraction
        if piston_pos is None:
            piston_pos = 1 / 3.0 * dashpot_length
        if piston_pos < 0:
            piston_pos = 0
        elif piston_pos > dashpot_length:
            piston_pos = dashpot_length - piston_thickness

        abs_piston_pos = p0.y + piston_pos

        gap = w * Dashpot._piston_gap_fraction
        shapes["piston"] = Composition(
            {
                "line": Line(p2, Point(B.x, abs_piston_pos + piston_thickness)),
                "rectangle": Rectangle(
                    Point(B.x - w + gap, abs_piston_pos),
                    2 * w - 2 * gap,
                    piston_thickness,
                ),
            }
        )
        shapes["piston"]["rectangle"].set_fill_pattern(Style.FillPattern.CROSS)

        super().__init__(shapes)

        # self.bar_length = s
        # self.width = 2 * w
        # self.piston_pos = piston_pos
        # self.dashpot_length = dashpot_length
        #
        # # Dimensions
        # start = Text_wArrow('start', (B[0]-1.5*w,B[1]-1.5*w), B)
        # width = Distance_wText((B[0]-w, B[1]-3.5*w), (B[0]+w, B[1]-3.5*w),
        #                        'width')
        # dplength = Distance_wText((B[0]+2*w, p0[1]), (B[0]+2*w, p1[1]),
        #                           'dashpot_length', text_pos=(B[0]+w,B[1]-w))
        # blength = Distance_wText((B[0]-2*w, B[1]), (B[0]-2*w, p0[1]),
        #                          'bar_length', text_pos=(B[0]-6*w,p0[1]-w))
        # ppos    = Distance_wText((B[0]-2*w, p0[1]), (B[0]-2*w, p0[1]+piston_pos),
        #                          'piston_pos', text_pos=(B[0]-6*w,p0[1]+piston_pos-w))
        # tlength = Distance_wText((B[0]+4*w, B[1]), (B[0]+4*w, B[1]+L),
        #                          'total_length',
        #                          text_pos=(B[0]+4.5*w, B[1]+L-2*w))
        # line = Line((B[0]+w, abs_piston_pos), (B[0]+7*w, abs_piston_pos))
        #   .set_linestyle('dashed')
        #   .set_linecolor('black')
        #   .set_linewidth(1)
        # pp = Text('abs_piston_pos', (B[0]+7*w, abs_piston_pos), alignment='left')
        # dims = {'start': start, 'width': width, 'dashpot_length': dplength,
        #         'bar_length': blength, 'total_length': tlength,
        #         'piston_pos': ppos,}
        # #'abs_piston_pos': Composition({'line': line, 'text': pp})}
        # self.dimensions = dims

    # def geometric_features(self):
    #     """
    #     Recorded geometric features:
    #     ==================== =============================================
    #     Attribute            Description
    #     ==================== =============================================
    #     start                Start point of dashpot.
    #     end                  End point of dashpot.
    #     bar_length           Length of first bar (from start to spring).
    #     dashpot_length       Length of dashpot middle part.
    #     width                Total width of dashpot.
    #     piston_pos           Position of piston in dashpot, relative to
    #                          start[1] + bar_length.
    #     ==================== =============================================
    #     """
    #     d = {'start': self.shapes['line start'].geometric_features()['start'],
    #          'end': self.shapes['piston']['line'].geometric_features()['start'],
    #          'bar_length': self.bar_length,
    #          'piston_pos': self.piston_pos,
    #          'width': self.width,
    #          'dashpot_length': self.dashpot_length,
    #          }
    #     return d



from typing import Dict

from pysketcher._arrow import Arrow
from pysketcher._line import Line
from pysketcher._point import Point
from pysketcher._shape import Shape
from pysketcher.annotation import LineAnnotation, TextPosition
from pysketcher.composition import Composition


class RadialDimension(Composition):
    """Used to indicate radial distances.

    Examples:
        >>> circle1 = ps.Circle(ps.Point(1.5, 1.5), 0.8)
        >>> circle1.style.line_width = 1
        >>> dim1 = RadialDimension(r"$r$", circle1.center, circle1(np.pi / 3))
        >>>
        >>> circle2 = ps.Circle(ps.Point(3.5, 1.5), 0.8)
        >>> circle2.style.line_width = 1
        >>> dim2 = RadialDimension(r"$r$", circle2.center, circle2(np.pi / 3))
        >>> dim2.diameter = True
        >>>
        >>> circle3 = ps.Circle(ps.Point(5.5, 1.5), 0.8)
        >>> circle3.style.line_width = 1
        >>> dim3 = RadialDimension(r"$r$", circle3.center, circle3(np.pi / 3))
        >>> dim3.center_mark = True
        >>>
        >>> circle4 = ps.Circle(ps.Point(7.5, 1.5), 0.8)
        >>> circle4.style.line_width = 1
        >>> dim4 = RadialDimension(r"$r$", circle4.center, circle4(np.pi / 3))
        >>> dim4.center_line = True
        >>>
        >>> circle5 = ps.Circle(ps.Point(9.5, 1.5), 0.8)
        >>> circle5.style.line_width = 1
        >>> dim5 = RadialDimension(r"$r$", circle5.center, circle5(np.pi / 3))
        >>> dim5.diameter = True
        >>> dim5.center_line = True
        >>>
        >>> fig = ps.Figure(0, 11, 0, 3, backend=MatplotlibBackend)
        >>> fig.add(circle1)
        >>> fig.add(dim1)
        >>> fig.add(circle2)
        >>> fig.add(dim2)
        >>> fig.add(circle3)
        >>> fig.add(dim3)
        >>> fig.add(circle4)
        >>> fig.add(dim4)
        >>> fig.add(circle5)
        >>> fig.add(dim5)
        >>> fig.save("pysketcher/images/radial_dimension.png")

    .. figure:: images/radial_dimension.png
        :alt: An example of LinearDimension.
        :figclass: align-center
        :scale: 30%

        An example of ``RadialDimension``.
    """

    _DEFAULT_DIAMETER: bool = False
    _DEFAULT_CENTER_LINE: bool = False
    _DEFAULT_CENTER_MARK: bool = False

    _DEFAULT_OFFSET: float = 0.2

    def __init__(self, text: str, center: Point, edge: Point):
        self._text = text
        self._center = center
        self._edge = edge
        self._offset = self._DEFAULT_OFFSET
        self._diameter = self._DEFAULT_DIAMETER
        self._center_line = self._DEFAULT_CENTER_LINE
        self._center_mark = self._DEFAULT_CENTER_MARK
        super().__init__(self._generate_shapes())

    def _generate_shapes(self) -> Dict[str, Shape]:
        arrow1 = Arrow(
            self._edge + (self._edge - self._center) * self._offset, self._edge
        )
        text = LineAnnotation(self._text, arrow1, TextPosition.START)
        text.style.font_size = 24
        ret_dict = {"arrow1": arrow1, "text": text}

        if self._diameter:
            inward_vector = self._center - self._edge
            offset_vector = inward_vector.unit_vector * self._offset
            start = self._edge + inward_vector * 2 + offset_vector
            end = self._edge + inward_vector * 2

            ret_dict["arrow2"] = Arrow(start, end)

        if self._center_mark:
            ret_dict["center_mark_h"] = Line(
                self._center - Point(-self._offset * 0.5, 0),
                self._center - Point(self._offset * 0.5, 0),
            )
            ret_dict["center_mark_v"] = Line(
                self._center - Point(0, -self._offset * 0.5),
                self._center - Point(0, self._offset * 0.5),
            )

        if self._center_line:
            ret_dict["center_line"] = Line(self._edge, self._center)

        if self._center_line and self._diameter:
            ret_dict["center_line2"] = Line(
                self._center, self._center + (self._center - self._edge)
            )
        return ret_dict

    @property
    def diameter(self) -> bool:
        """If true, the dimension indicates diameter rather than radius."""
        return self._diameter

    @diameter.setter
    def diameter(self, diameter: bool):
        self._diameter = diameter
        self._shapes = self._generate_shapes()

    @property
    def center_mark(self) -> bool:
        """If true, a center mark will be drawn."""
        return self._center_mark

    @center_mark.setter
    def center_mark(self, center_mark: bool):
        self._center_mark = center_mark
        self._shapes = self._generate_shapes()

    @property
    def center_line(self) -> bool:
        """If true then a line will be drawn from the center to the circumference."""
        return self._center_line

    @center_line.setter
    def center_line(self, center_line: bool):
        self._center_line = center_line
        self._shapes = self._generate_shapes()



"""Provide a means to dimension shapes."""
from pysketcher.dimension._angular_dimension import AngularDimension
from pysketcher.dimension._linear_dimension import LinearDimension
from pysketcher.dimension._radial_dimension import RadialDimension

__all__ = ["AngularDimension", "LinearDimension", "RadialDimension"]



from pysketcher._curve import Curve
from pysketcher._point import Point


class Triangle(Curve):
    """Triangle defined by its three vertices p1, p2, and p3.

    Args:
        p1: The first ``Point``.
        p2: The second ``Point``.
        p3: The third ``Point``.

    Examples:
        >>> model = ps.Triangle(ps.Point(1, 1), ps.Point(1, 4), ps.Point(3, 3))
        >>> fig = ps.Figure(0, 5, 0, 5, backend=MatplotlibBackend)
        >>> fig.add(model)
        >>> fig.save("pysketcher/images/triangle.png")

        .. figure:: images/triangle.png
            :alt: An example of Triangle.
            :figclass: align-center
            :scale: 30%

            An example of ``Triangle``.
    """

    def __init__(self, p1: Point, p2: Point, p3: Point):
        self._p1 = p1
        self._p2 = p2
        self._p3 = p3

        super().__init__([p1, p2, p3, p1])

    def rotate(self, angle: float, center: Point):
        """Rotates the Triangle through a ``angle`` about ``center``.

        Args:
            angle: The angle through which the triangle should be rotated in radians.
            center: The point about which the triangle should be rotated.

        Returns:
            A copy of the triangle subjected to the specified rotation.
        """
        return Triangle(
            self._p1.rotate(angle, center),
            self._p2.rotate(angle, center),
            self._p3.rotate(angle, center),
        )



from collections import namedtuple
from enum import auto, unique

from pysketcher._utils import DocEnum

Arrow = namedtuple("Arrow", "start end")


class Style:
    """Represents the visual characteristics of a Drawable object."""

    @unique
    class LineStyle(DocEnum):
        """Represents the manner of line to be drawn."""

        SOLID = auto(), "An unbroken line."
        DOTTED = auto(), "A dotted line."
        DASH_DOT = auto(), "An alternating dash dot line."
        DASHED = auto(), "A dashed line."

    @unique
    class Color(DocEnum):
        """Represents the color in which something should be rendered."""

        GREY = auto(), "The color grey."
        BLACK = auto(), "The color black."
        BROWN = auto(), "The color brown."
        RED = auto(), "The color red."
        ORANGE = auto(), "The color orange."
        YELLOW = auto(), " The color yellow."
        GREEN = auto(), "The color green."
        CYAN = auto(), "The color cyan."
        BLUE = auto(), "The color blue."
        MAGENTA = auto(), "The color magenta."
        PURPLE = auto(), "The color purple."
        WHITE = auto(), "The color white."

    @unique
    class FillPattern(DocEnum):
        """Represents the pattern in which something can be shaded."""

        VERTICAL = auto(), "A pattern made up for vertical lines."
        HORIZONTAL = auto(), "A pattern made up of horizontal lines."
        CROSS = auto(), "A pattern made of crossing diagonal lines."
        SQUARE = auto(), "A pattern made of crossing horizontal and vertical lines."
        STAR = auto(), "A pattern made of stars."
        DOT = auto(), "A pattern made of dots,"
        CIRCLE = auto(), "A pattern made of circles."
        SMALL_CIRCLE = auto(), "A pattern made of small circles."
        UP_RIGHT_TO_LEFT = auto(), "A pattern made of diagonal lines from SE to NW."
        UP_LEFT_TO_RIGHT = auto(), "A pattern made of diagonal lines from SW to NE."

    @unique
    class ArrowStyle(DocEnum):
        """Represents the style of arrow which can adorn a line."""

        START = Arrow(True, False), "Place an arrowhead at the start of the line."
        END = Arrow(False, True), "Place an arrowhead at the end of the line."
        DOUBLE = Arrow(True, True), "Place an arrowhead at both ends of the line."

    _line_style: LineStyle
    _line_width: float
    _line_color: Color
    _fill_pattern: FillPattern
    _fill_color: Color
    _arrow: ArrowStyle
    _shadow: int

    def __init__(self):
        self._line_style = self.LineStyle.SOLID
        self._line_width = 1
        self._line_color = self.Color.BLACK
        self._fill_pattern = None
        self._fill_color = None
        self._arrow = None
        self._shadow = None

    @property
    def line_style(self) -> LineStyle:
        """The style of line."""
        return self._line_style

    @line_style.setter
    def line_style(self, line_style: LineStyle):
        self._line_style = line_style

    @property
    def line_width(self) -> float:
        """The width of line."""
        return self._line_width

    @line_width.setter
    def line_width(self, line_width: float):
        self._line_width = line_width

    @property
    def line_color(self) -> Color:
        """The color of line."""
        return self._line_color

    @line_color.setter
    def line_color(self, color: Color):
        self._line_color = color

    @property
    def fill_color(self) -> Color:
        """The color of fill."""
        return self._fill_color

    @fill_color.setter
    def fill_color(self, color: Color):
        self._fill_color = color

    @property
    def fill_pattern(self) -> FillPattern:
        """The pattern of fill."""
        return self._fill_pattern

    @fill_pattern.setter
    def fill_pattern(self, fill_pattern: FillPattern):
        self._fill_pattern = fill_pattern
        if not self.fill_color:
            self.fill_color = self.Color.WHITE

    @property
    def arrow(self) -> ArrowStyle:
        """The arrow which should adorn the line."""
        return self._arrow

    @arrow.setter
    def arrow(self, arrow: ArrowStyle):
        self._arrow = arrow

    @property
    def shadow(self) -> float:
        """The shadow which the object should cast."""
        return self._shadow

    @shadow.setter
    def shadow(self, shadow: float):
        self._shadow = shadow

    def __str__(self):
        """Returns a representation of the style in string form."""
        return (
            "line_style: %s, line_width: %s, line_color: %s,"
            " fill_pattern: %s, fill_color: %s, arrow: %s shadow: %s"
            % (
                self.line_style,
                self.line_width,
                self.line_color,
                self.fill_pattern,
                self.fill_color,
                self.arrow,
                self.shadow,
            )
        )


class TextStyle(Style):
    """Represents the visual style of a text object."""

    class FontFamily(DocEnum):
        """Represents the font which should be used to render text."""

        SERIF = auto(), "A font with serifs."
        SANS = auto(), "A font without serifs."
        MONO = auto(), "An equally spaced font."

    class Alignment(DocEnum):
        """Represents the horizontal alignment of text."""

        LEFT = auto(), "Align text to the left."
        CENTER = auto(), "Align text to the center."
        RIGHT = auto(), "Align text to the right."

    _font_size: float
    _font_family: FontFamily
    _alignment: Alignment

    def __init__(
        self,
    ):
        super().__init__()
        self._font_size = 18
        self._font_family = self.FontFamily.SANS
        self._alignment = self.Alignment.CENTER

    @property
    def font_size(self) -> float:
        """The size of text."""
        return self._font_size

    @font_size.setter
    def font_size(self, font_size: float):
        self._font_size = font_size

    @property
    def font_family(self) -> FontFamily:
        """The family of font in which text is rendered."""
        return self._font_family

    @font_family.setter
    def font_family(self, font_family: FontFamily):
        self._font_family = font_family

    @property
    def alignment(self):
        """The alignment to be used when text is rendered."""
        return self._alignment

    @alignment.setter
    def alignment(self, alignment: Alignment):
        self._alignment = alignment



import math
from typing import Callable, List, Tuple

import numpy as np

from pysketcher._curve import Curve
from pysketcher._point import Point


class CubicBezier(Curve):
    """A cubic bezier curve implementation.

    Examples:
        >>> s = ps.CubicBezier(
        ...     ps.Point(0, 0), [(ps.Point(1, 1), ps.Point(2.5, 0.5), ps.Point(2, 2.5))]
        ... )
        >>> fig = ps.Figure(-0.5, 3.0, -0.5, 3.0, backend=MatplotlibBackend)
        >>> fig.add(s)
        >>> fig.save("pysketcher/images/cubic_bezier.png")

        .. figure:: images/cubic_bezier.png
            :alt: An example of a CubicBezier.
            :figclass: align-center

            An example of ``CubicBezier``.
    """

    _input_points: List[Tuple[Point, Point, Point, Point]]
    _segments: List[Callable[[float], Point]]
    _points: List[Point]

    def __init__(
        self,
        start: Point,
        points: List[Tuple[Point, Point, Point]],
        resolution: int = 50,
    ):
        start_point = start
        self._input_points = []
        self._segments = []
        for pts in points:
            self._input_points.append((start_point,) + pts)
            self._segments.append(self._segment(start_point, pts[0], pts[1], pts[2]))
            start_point = pts[2]

        ts = np.linspace(0, len(self._segments), resolution, endpoint=False)
        self._points = [self.__call__(t) for t in ts]
        last_point = points[len(points) - 1][2]
        self._points.append(last_point)
        super().__init__(self._points)

    @staticmethod
    def _segment(
        p0: Point, p1: Point, p2: Point, p3: Point
    ) -> Callable[[float], Point]:
        def _bernstein_cubic(a: List[float], t: float) -> float:
            return (
                (1 - t) ** 3 * a[0]
                + (1 - t) ** 2 * 3 * t * a[1]
                + (1 - t) * 3 * t ** 2 * a[2]
                + t ** 3 * a[3]
            )

        def _segment_function(t: float) -> Point:
            x = _bernstein_cubic([p0.x, p1.x, p2.x, p3.x], t)
            y = _bernstein_cubic([p0.y, p1.y, p2.y, p3.y], t)
            return Point(x, y)

        return _segment_function

    @property
    def start(self) -> Point:
        """The first point of the curve."""
        return self._points[0]

    @property
    def end(self) -> Point:
        """The last point of the curve."""
        return self._points[-1]

    def __call__(self, t: float) -> Point:
        """Given a parameter, ``t``, returns the point on the curve."""
        t, index = math.modf(t)
        if index > len(self._segments) or t < 0:
            raise ValueError(
                (
                    f"Value of t is only valid > 0  and < {len(self._segments)} "
                    f"got {t}."
                )
            )
        return self._segments[int(index)](t)
        pass



from matplotlib.axes import Axes

from pysketcher.backend.matplotlib._matplotlib_adapter import MatplotlibAdapter
from pysketcher.composition import Composition


class MatplotlibComposition(MatplotlibAdapter):
    def __init__(self, mplb):
        self._mplb = mplb

    def plot(self, shape: Composition, axes: Axes):
        shape.apply(self._mplb.add)



from pysketcher._line import Line
from pysketcher._point import Point
from pysketcher._style import TextStyle
from pysketcher._text import Text
from pysketcher.annotation._text_position import TextPosition


class LineAnnotation(Text):
    """Annotates a line with the provided text."""

    # TODO: Write a LineAnnotation Example
    _DEFAULT_SPACING: Point = Point(0.15, 0.15)

    def __init__(
        self, text: str, line: Line, text_position: TextPosition = TextPosition.MIDDLE
    ):

        spacing = self._DEFAULT_SPACING

        if text_position == TextPosition.START:
            position = line.start + spacing
            alignment = TextStyle.Alignment.LEFT
        elif text_position == TextPosition.END:
            position = line.end + spacing
            alignment = TextStyle.Alignment.RIGHT
        elif text_position == TextPosition.MIDDLE:
            position = line.start + (line.end - line.start) * 0.5 + spacing
            alignment = TextStyle.Alignment.CENTER
        else:
            raise RuntimeError(f"Invalid value of text_position: {text_position}.")

        super().__init__(text, position)
        self.style.alignment = alignment



from typing import List

import numpy as np

from pysketcher._angle import Angle
from pysketcher._point import Point
from pysketcher._shape import Shape


class Curve(Shape):
    """General curve as a sequence of (x,y) coordintes.

    Examples:
        >>> code = ps.Curve(
        ...     [
        ...         ps.Point(0, 0),
        ...         ps.Point(1, 1),
        ...         ps.Point(2, 4),
        ...         ps.Point(3, 9),
        ...         ps.Point(4, 16),
        ...     ]
        ... )
        >>> code.style.line_color = ps.Style.Color.BLACK
        >>> model = ps.Composition(dict(text=code))
        >>> fig = ps.Figure(0, 5, 0, 16, backend=MatplotlibBackend)
        >>> fig.add(model)
        >>> fig.save("pysketcher/images/curve.png")


    .. figure:: images/curve.png
        :alt: An example of Curve.
        :figclass: align-center
        :scale: 30%

        An example of ``Curve``.
    """

    def __init__(self, points: List[Point]):
        super().__init__()
        self._points = points
        # self.shapes must not be defined in this class
        # as self.shapes holds children objects:
        # Curve has no children (end leaf of self.shapes tree)

    @property
    def points(self):
        """The points which make up the curve."""
        return self._points

    @property
    def xs(self):
        """The x co-ordinates of the points of the curve."""
        return np.array([p.x for p in self.points])

    @property
    def ys(self):
        """The y co-ordinates of the Points of the curve."""
        return np.array([p.y for p in self.points])

    def rotate(self, angle: Angle, center: Point) -> "Curve":
        """Rotate all coordinates.

        Args:
            angle: The ``Angle`` (in radians) through which the rotation should be
                performed.
            center: The ``Point`` about which the rotation should be performed.

        Returns:
            A copy of the ``Curve`` which has had all points rotated in the
                described manner.
        """
        ret_curve = Curve([p.rotate(angle, center) for p in self.points])
        ret_curve.style = self.style
        return ret_curve

    def scale(self, factor: float) -> "Curve":
        """Scale all coordinates by `factor`: ``x = factor*x``, etc."""
        ret_curve = Curve(
            Point.from_coordinate_lists(factor * self.xs, factor * self.ys)
        )
        ret_curve.style = self.style
        return ret_curve

    def translate(self, vec: Point) -> "Curve":
        """Translate all coordinates by a vector `vec`."""
        ret_curve = Curve([p + vec for p in self.points])
        ret_curve.style = self.style
        return ret_curve

    def __str__(self):
        """Compact pretty print of a Curve object."""
        s = "%d (x,y) coords" % len(self._points)
        return s

    def __repr__(self):
        """An unambiguous string representational of a Curve object."""
        return "Curve(" + ",".join([repr(p) for p in self.points]) + ")"



from pysketcher._point import Point
from pysketcher._rectangle import Rectangle
from pysketcher._triangle import Triangle
from pysketcher.composition import Composition


class SimpleSupport(Composition):
    """A representation of a simple support.

    Often used in static load analysis, this shows a diagrammatic
    representation of a point support.

    Args:
         position: The top of the simple support.
         size: The distance from the top of the simple support to the center of
            the base.

    Examples:
        >>> s = ps.SimpleSupport(ps.Point(1.0, 1.0), 0.5)
        >>> fig = ps.Figure(0, 2.0, 0, 1.5, backend=MatplotlibBackend)
        >>> fig.add(s)
        >>> fig.save("pysketcher/images/simple_support.png")

        .. figure:: images/simple_support.png
            :alt: An example of a SimpleSupport.
            :figclass: align-center

            An example of ``SimpleSupport``.

    """

    _position: Point
    _size: float

    def __init__(self, position: Point, size: float):
        self._position = position
        self._size = size
        self._p0 = Point(position.x - size / 2.0, position.y - size)
        self._p1 = Point(position.x + size / 2.0, position.y - size)
        self._triangle = Triangle(self._p0, self._p1, position)
        gap = size / 5.0
        self._height = size / 4.0  # height of rectangle
        self._p2 = Point(self._p0.x, self._p0.y - gap - self._height)
        self._rectangle = Rectangle(self._p2, self._size, self._height)
        shapes = {"triangle": self._triangle, "rectangle": self._rectangle}
        super().__init__(shapes)

    @property
    def mid_support(self) -> Point:
        """Returns the midpoint of the base of the support."""
        return (self._rectangle.lower_left + self._rectangle.lower_right) * 0.5



from typing import List

from pysketcher import Angle
from pysketcher._curve import Curve
from pysketcher._point import Point


class Rectangle(Curve):
    """A representation of a rectangle.

    Rectangle specified by the point `lower_left_corner`, `width`,
    and `height`.

    Args:
        lower_left_corner: The point from which the ``Rectangle`` should
            be drawn.
        width: The width of the ``Rectangle``.
        height: The height of the ``Rectangle``.

    Examples:
        >>> code = ps.Rectangle(ps.Point(1, 1), 3, 4)
        >>> model = ps.Composition(dict(text=code))
        >>> fig = ps.Figure(0, 5, 0, 5, backend=MatplotlibBackend)
        >>> fig.add(model)
        >>> fig.save("pysketcher/images/rectangle.png")


    .. figure:: images/rectangle.png
        :alt: An example of Rectangle.
        :figclass: align-center
        :scale: 50%

        An example of ``Rectangle``.
    """

    def __init__(self, lower_left_corner: Point, width: float, height: float):

        self._width = width
        self._height = height
        self._lower_left_corner = lower_left_corner

        super().__init__(self._generate_points())

    def _generate_points(self) -> List[Point]:
        return [
            self._lower_left_corner,
            self._lower_left_corner + Point(self._width, 0),
            self._lower_left_corner + Point(self._width, self._height),
            self._lower_left_corner + Point(0, self._height),
            self._lower_left_corner,
        ]

    def rotate(self, angle: Angle, center: Point) -> Curve:
        """Rotates the rectangle through ``angle`` radians about ``center``.

        Args:
            angle: The angle in radians through which the rotation should be.
            center: The point about which the rotation should be performed.

        Returns:
            A curve which looks like a rectangle rotated. We need to fix this
            so that the rectangle class supports rectangles with sides which are not
            parallel to the horizontal.
        """
        points = []
        for point in self._points:
            points.append(point.rotate(angle, center))
        return Curve(points)

    @property
    def width(self) -> float:
        """The width of the rectangle."""
        return self._width

    @property
    def height(self) -> float:
        """The height of the rectangle."""
        return self._height

    @property
    def lower_left(self) -> Point:
        """The lower left point of the rectangle."""
        return self._lower_left_corner

    @property
    def lower_right(self) -> Point:
        """The lower right point of the rectangle."""
        return self._lower_left_corner + Point(self._width, 0)

    @property
    def upper_left(self) -> Point:
        """The upper left point of the rectangle."""
        return self._lower_left_corner + Point(0, self._height)

    @property
    def upper_right(self) -> Point:
        """The upper right point of the rectangle."""
        return self._lower_left_corner + Point(self._width, self._height)



from typing import List

import numpy as np
from scipy.interpolate import UnivariateSpline

from pysketcher._curve import Curve
from pysketcher._point import Point


class Spline(Curve):
    """A univariate spline.

    Note: UnivariateSpline interpolation may not work if
        the x[i] points are far from uniformly spaced.

    Examples:
        >>> s = ps.Spline(
        ...     [
        ...         ps.Point(0, 0),
        ...         ps.Point(1, 1),
        ...         ps.Point(2, 4),
        ...         ps.Point(3, 9),
        ...         ps.Point(4, 16),
        ...     ]
        ... )
        >>> fig = ps.Figure(0, 5, 0, 16, backend=MatplotlibBackend)
        >>> fig.add(s)
        >>> fig.save("pysketcher/images/spline.png")

        .. figure:: images/spline.png
            :alt: An example of a Spline.
            :figclass: align-center
            :scale: 30%

            An example of ``Spline``.
    """

    _input_points: List[Point]
    _smooth: UnivariateSpline

    def __init__(self, points: List[Point], degree: int = 3, resolution: int = 501):
        self._input_points = points
        self._smooth = UnivariateSpline(
            [p.x for p in points], [p.y for p in points], s=0, k=degree
        )
        x_coordinates = np.linspace(points[0].x, points[-1].x, resolution)
        y_coordinates = self._smooth(x_coordinates)
        smooth_points = [Point(p[0], p[1]) for p in zip(x_coordinates, y_coordinates)]
        super().__init__(smooth_points)

    def __call__(self, x):
        """Returns the value of the curve at a given x-coordinate."""
        return self._smooth(x)



import matplotlib.pyplot as plt

from pysketcher._text import Text
from pysketcher.backend.matplotlib._matplotlib_adapter import MatplotlibAdapter
from pysketcher.backend.matplotlib._matplotlib_style import MatplotlibTextStyle


class MatplotlibText(MatplotlibAdapter):
    """Renders a Text primitive."""

    def plot(self, text: Text, axes: plt.Axes):
        """Render a Text primitive.

        Write `text` string at a position (centered, left, right - according
        to the `alignment` string). `position` is a point in the coordinate
        system.

        Args:
            text: the ``Text`` object to be rendered.
            axes: the ``Axes`` to render the text object on.
        """
        mpl_style = MatplotlibTextStyle(text.style)
        kwargs = {}
        if mpl_style.font_family is not None:
            kwargs["family"] = mpl_style.font_family
        if mpl_style.fill_color is not None:
            kwargs["backgroundcolor"] = mpl_style.fill_color
        if mpl_style.line_color is not None:
            kwargs["color"] = mpl_style.line_color

        rotation_angle = text.direction.angle
        if rotation_angle != 0.0:
            kwargs["rotation"] = rotation_angle

        axes.text(
            text.position.x,
            text.position.y,
            text.text,
            horizontalalignment=mpl_style.alignment,
            fontsize=mpl_style.font_size,
            **kwargs
        )



from enum import auto, Enum, unique
from typing import Dict

from pysketcher._arrow import DoubleArrow
from pysketcher._line import Line
from pysketcher._point import Point
from pysketcher._shape import Shape
from pysketcher.annotation import LineAnnotation, TextPosition
from pysketcher.composition import Composition


class LinearDimension(Composition):
    """Used to indicate the linear distance between two points.

    Examples:
        >>> triangle1 = ps.Triangle(ps.Point(1, 1), ps.Point(1, 2), ps.Point(3, 1))
        >>> dim1 = LinearDimension(r"$c$", ps.Point(3, 1), ps.Point(1, 2))
        >>> text1 = ps.Text("Aligned", ps.Point(2, 0.5))
        >>>
        >>> triangle2 = ps.Triangle(ps.Point(5, 1), ps.Point(5, 2), ps.Point(7, 1))
        >>> dim2 = LinearDimension(r"$a$", ps.Point(7, 1), ps.Point(5, 2))
        >>> dim2.offset_style = LinearDimension.OffsetStyle.HORIZONTAL
        >>> text2 = ps.Text("Horizontal", ps.Point(6, 0.5))
        >>>
        >>> triangle3 = ps.Triangle(ps.Point(9, 1), ps.Point(9, 2), ps.Point(11, 1))
        >>> dim3 = LinearDimension(r"$b$", ps.Point(11, 1), ps.Point(9, 2))
        >>> dim3.offset_style = LinearDimension.OffsetStyle.VERTICAL
        >>> text3 = ps.Text("Vertical", ps.Point(10, 0.5))
        >>>
        >>> fig = ps.Figure(0, 13, 0, 3, backend=MatplotlibBackend)
        >>> fig.add(triangle1)
        >>> fig.add(dim1)
        >>> fig.add(text1)
        >>> fig.add(triangle2)
        >>> fig.add(dim2)
        >>> fig.add(text2)
        >>> fig.add(triangle3)
        >>> fig.add(dim3)
        >>> fig.add(text3)
        >>> fig.save("pysketcher/images/linear_dimension.png")

        .. figure:: images/linear_dimension.png
            :alt: An example of LinearDimension.
            :figclass: align-center
            :scale: 30%

            An example of ``LinearDimension``.
    """

    _DEFAULT_OFFSET: float = 0.5
    _DEFAULT_MINOR_OFFSET: float = 0.1

    @unique
    class OffsetStyle(Enum):
        """The style of offset in a Linear dimension."""

        HORIZONTAL = auto()
        VERTICAL = auto()
        ALIGNED = auto()

    def __init__(
        self,
        text: str,
        start: Point,
        end: Point,
        offset_style: OffsetStyle = OffsetStyle.ALIGNED,
    ):
        self._start = start
        self._end = end
        self._text = text
        self._offset_style = offset_style
        self._offset = self._DEFAULT_OFFSET
        self._minor_offset = self._DEFAULT_MINOR_OFFSET

        super().__init__(self._generate_shapes())

    def _generate_shapes(self) -> Dict[str, Shape]:
        if self._offset_style == self.OffsetStyle.ALIGNED:
            self._dimension_vector = self._start - self._end
            self._offset_vector = self._dimension_vector.normal * self.offset

            arrow = DoubleArrow(
                self._start + self._offset_vector, self._end + self._offset_vector
            )

        elif self._offset_style == self.OffsetStyle.HORIZONTAL:
            arrow = self._horizontal_arrow()

        elif self._offset_style == self.OffsetStyle.VERTICAL:
            arrow = self._vertical_arrow()
        else:
            raise NotImplementedError()

        annotation = LineAnnotation(self._text, arrow, TextPosition.MIDDLE)

        def extension_line(start: Point, end: Point) -> Line:
            return Line(
                start + (end - start) * self._minor_offset,
                end + (end - start) * self._minor_offset,
            )

        return {
            "arrow": arrow,
            "annotation": annotation,
            "extension_line_1": extension_line(self._start, arrow.start),
            "extension_line_2": extension_line(self._end, arrow.end),
        }

    def _horizontal_arrow(self) -> DoubleArrow:
        if self._start.x < self._end.x:
            return DoubleArrow(
                Point(self._start.x, self._start.y + self._offset),
                Point(self._end.x, self._start.y + self._offset),
            )
        else:
            return DoubleArrow(
                Point(self._start.x, self._end.y + self._offset),
                Point(self._end.x, self._end.y + self._offset),
            )

    def _vertical_arrow(self) -> DoubleArrow:
        if self._start.y > self._end.y:
            return DoubleArrow(
                Point(self._end.x + self._offset, self._start.y),
                Point(self._end.x + self._offset, self._end.y),
            )
        else:
            return DoubleArrow(
                Point(self._start.x + self._offset, self._start.y),
                Point(self._start.x + self._offset, self._end.y),
            )

    @property
    def offset_style(self) -> OffsetStyle:
        """Specifies the orientation of the drawn dimension."""
        return self._offset_style

    @offset_style.setter
    def offset_style(self, offset_style: OffsetStyle):
        self._offset_style = offset_style
        self._shapes = self._generate_shapes()

    @property
    def offset(self) -> float:
        """How far the dimension should be offset from the provided points."""
        return self._offset

    @offset.setter
    def offset(self, offset: float):
        self._offset = offset
        self._shapes = self._generate_shapes()

    @property
    def start(self) -> Point:
        """The start of the dimension."""
        return self._start

    @property
    def end(self) -> Point:
        """The end of the dimension."""
        return self._end



from typing import Callable

from pysketcher._arrow import Arrow
from pysketcher._line import Line
from pysketcher._point import Point
from pysketcher._spline import Spline
from pysketcher.composition import Composition


class VelocityProfile(Composition):
    """A representation of the profile of velocity in laminar flow.

    Args:
        start: the point from which the profile should start.
        height: the height of the profile.
        profile: a function which provides the value of the profile at a given point
        num_arrows: the number of arrows to display
        scaling: a scaling factor

    Examples:
        >>> def velocity_profile(y: float) -> ps.Point:
        ...     return ps.Point(y * (8 - y) / 4, 0)
        >>> pr = ps.VelocityProfile(ps.Point(0, 0), 4, velocity_profile, 5)
        >>> fig = ps.Figure(0, 4.1, 0, 4, backend=MatplotlibBackend)
        >>> fig.add(pr)
        >>> fig.save("pysketcher/images/velocity_profile.png")

        .. figure:: images/velocity_profile.png
            :alt: An example of a Velocity Profile.
            :figclass: align-center
            :scale: 50%

            An example of ``VelocityProfile``.
    """

    _start: Point
    _height: float
    _profile: Callable[[float], Point]
    _num_arrows: int
    _scaling: float

    def __init__(
        self,
        start: Point,
        height: float,
        profile: Callable[[float], Point],
        num_arrows: int,
        scaling: float = 1,
    ):

        self._start = start
        self._height = height
        self._profile = profile
        self._num_arrows = num_arrows
        self._scaling = scaling

        shapes = dict()

        # Draw left line
        shapes["start line"] = Line(self._start, (self._start + Point(0, self._height)))

        # Draw velocity arrows
        dy = float(self._height) / (self._num_arrows - 1)

        end_points = []

        for i in range(self._num_arrows):
            start_position = Point(start.x, start.y + i * dy)
            end_position = start_position + profile(start_position.y) * self._scaling
            end_points += [end_position]
            if start_position == end_position:
                continue
            shapes["arrow%d" % i] = Arrow(start_position, end_position)

        shapes["smooth curve"] = Spline(end_points)
        super().__init__(shapes)



import logging
from typing import Callable, Dict, Optional, Tuple, Type, Union

from celluloid import Camera
from matplotlib.animation import ArtistAnimation
import matplotlib.pyplot as plt

from pysketcher._curve import Curve
from pysketcher._drawable import Drawable
from pysketcher._text import Text
from pysketcher.backend.backend import Backend
from pysketcher.backend.matplotlib._matplotlib_adapter import MatplotlibAdapter
from pysketcher.backend.matplotlib._matplotlib_composition import MatplotlibComposition
from pysketcher.backend.matplotlib._matplotlib_curve import MatplotlibCurve
from pysketcher.backend.matplotlib._matplotlib_text import MatplotlibText
from pysketcher.composition import Composition


# plt.rc("text", usetex=True)
# plt.rcParams["text.latex.preamble"] = r"\usepackage{amsmath}"


class MatplotlibBackend(Backend):
    """Simple interface for plotting. Makes use of Matplotlib for plotting."""

    _fig: plt.Figure
    _axes: plt.Axes
    _camera: Optional[Camera]
    _x_min: float
    _y_min: float
    _x_max: float
    _y_max: float

    _INTERVAL: int = 40  # ms between each frame

    def __init__(self, x_min, x_max, y_min, y_max):
        plt.ion()
        self._x_min = x_min
        self._x_max = x_max
        self._y_min = y_min
        self._y_max = y_max
        self._camera = None
        self._fig = plt.figure(
            figsize=[(x_max - x_min) * 3, (y_max - y_min) * 3], tight_layout=False
        )
        self._axes = self._fig.gca()
        self._configure_axes()

    def _configure_axes(self):
        self._axes.set_xlim(self._x_min, self._x_max)
        self._axes.set_ylim(self._y_min, self._y_max)
        self._axes.set_aspect("equal")
        self._axes.set_axis_off()

    def add(self, shape: Drawable) -> None:
        for typ, adapter in self._adapters.items():
            if issubclass(shape.__class__, typ):
                adapter.plot(shape, self._axes)

    def erase(self):
        self._fig.clear()
        self._axes = self._fig.gca()
        self._configure_axes()

    def show(self):
        self._fig.sca(self._axes)
        self._fig.canvas.draw()
        self._fig.show()

    def save(self, filename: str) -> None:
        logging.info(f"Saving to {filename}.")
        # TODO: manage formats
        self._fig.savefig(filename)

    @property
    def _adapters(self) -> Dict[Type, MatplotlibAdapter]:
        return {
            Curve: MatplotlibCurve(),
            Text: MatplotlibText(),
            Composition: MatplotlibComposition(self),
        }

    def animate(
        self,
        func: Callable[[float], Drawable],
        interval: Union[Tuple[float, float], Tuple[float, float, float]],
    ):
        if len(interval) == 2:
            start, end = interval
            increment = 1
        else:
            start, end, increment = interval
        self._camera = Camera(self._fig)
        i: float = start
        while i < end:
            self.add(func(i))
            self._camera.snap()
            i += increment

    def show_animation(self):
        animation: ArtistAnimation = self._camera.animate(interval=self._INTERVAL)
        animation.show()

    def save_animation(self, filename: str):
        animation: ArtistAnimation = self._camera.animate(interval=self._INTERVAL)
        animation.save(filename)



from pysketcher._arrow import Arrow
from pysketcher._point import Point
from pysketcher._style import Style
from pysketcher.annotation import LineAnnotation, TextPosition
from pysketcher.composition import Composition


class Force(Composition):
    """A composition of Arrow and LinearAnnotation."""

    def __init__(
        self,
        text: str,
        start: Point,
        end: Point,
        text_position: TextPosition = TextPosition.MIDDLE,
    ):
        self._text = text
        self._start = start
        self._end = end

        self._arrow = Arrow(start, end)
        self._label = LineAnnotation(text, self._arrow, text_position)
        super().__init__({"arrow": self._arrow, "label": self._label})


class Gravity(Force):
    """Downward-pointing gravity arrow with the symbol g."""

    # TODO: add the g

    def __init__(
        self,
        start,
        length,
    ):
        Force.__init__(self, "$g$", start, start + Point(0, -length))
        self.style.line_color = Style.Color.BLACK



import numpy as np


class Angle(float):
    r"""A representation of a 2-Dimensional Angle.

    An ``Angle`` is a float which is limited in value to between :math:`\pi`
    and :math:`-\pi`. Values outside of this range have :math:`2\pi` either added
    or subtracted repeatedly until they are within this bound.

    Args:
        value: The value of the angle to create.

    Examples:
        >>> Angle(1.0)
        Angle(1.0)

        >>> Angle(np.pi / 2)
        Angle(1.5707963267948966)

        >>> Angle(123)
        Angle(-2.6637061435917175)

        >>> Angle(2 * np.pi)
        Angle(0.0)
    """

    def __new__(cls, value: float) -> "Angle":
        """Creates a new Angle."""
        value = cls._normalize(value)
        return super(cls, cls).__new__(cls, value)

    @staticmethod
    def _normalize(value: float):
        def order(x: float) -> float:
            return np.trunc(x / (2 * np.pi))

        if not (-np.pi < value <= np.pi):
            order_value = order(value)
            value = value - order_value * 2 * np.pi
            # TODO: for some reason, ``order`` sometimes overshoots. Need to fix this.
            while value <= -np.pi:
                value = value + 2 * np.pi
            while value > np.pi:
                value = value - 2 * np.pi
        return value

    def __add__(self, other: float) -> "Angle":
        """Returns the sum of this ``Angle`` with the provided ``Angle``."""
        res = super(self.__class__, self).__add__(other)
        return self.__class__(res)

    def __sub__(self, other: float) -> "Angle":
        """Returns the difference between this ``Angle`` and the provided ``Angle``."""
        res = super(self.__class__, self).__sub__(other)
        return self.__class__(res)

    def __mul__(self, other: float) -> "Angle":
        """Scales the ``Angle`` by a factor of the provided value."""
        res = super(self.__class__, self).__mul__(other)
        return self.__class__(res)

    def __truediv__(self, other: float) -> "Angle":
        """Divides the ``Angle`` by the provided factor."""
        res = super(self.__class__, self).__truediv__(other)
        return self.__class__(res)

    def __str__(self: float) -> str:
        """Provides a string rendition of the value of the ``Angle``."""
        return f"{float(self)}"

    def __repr__(self: float) -> str:
        """Provides a string rendition of the value of the ``Angle`` and class."""
        return f"Angle({float(self)})"



from typing import Callable, Dict, TypeVar, Union

from pysketcher._point import Point
from pysketcher._shape import Shape, Stylable
from pysketcher._style import Style
from pysketcher._text import Text

T = TypeVar("T")


class Composition(Stylable):
    """Provides a convenient means to group shapes together."""

    class CompositionStyle(Style):
        """Presents the Stylable contract for a Composition.

        Sets the style of each object in the composition
        transparently.
        """

        composition: "Composition"

        def __init__(self, composition: "Composition"):
            super().__init__()
            self.composition = composition

        @property
        def line_style(self) -> Style.LineStyle:
            """The line style of all ``Shapes`` in the composition."""
            return super().line_style

        @line_style.setter
        def line_style(self, line_style: Style.LineStyle):
            for shape in self.composition:
                shape.style.line_style = line_style

        @property
        def line_width(self) -> float:
            """The line width of all ``Shapes`` in the composition."""
            return super().line_width

        @line_width.setter
        def line_width(self, line_width: float):
            for shape in self.composition:
                shape.style.line_width = line_width

        @property
        def line_color(self) -> Style.Color:
            """The line color of all ``Shapes`` in the composition."""
            return super().line_color

        @line_color.setter
        def line_color(self, line_color: Style.Color):
            for shape in self.composition:
                shape.style.line_color = line_color

        @property
        def fill_color(self) -> Style.Color:
            """The fill color of all ``Shapes`` in the composition."""
            return super().fill_color

        @fill_color.setter
        def fill_color(self, fill_color: Style.Color):
            for shape in self.composition:
                shape.style.fill_color = fill_color

        @property
        def fill_pattern(self) -> Style.FillPattern:
            """The fill pattern of all ``Shapes`` in the composition."""
            return super().fill_pattern

        @fill_pattern.setter
        def fill_pattern(self, fill_pattern: Style.Color):
            for shape in self.composition:
                shape.style.fill_pattern = fill_pattern

        @property
        def arrow(self) -> Style.ArrowStyle:
            """The arrow nature of all ``Shapes`` in the composition."""
            return super().arrow

        @arrow.setter
        def arrow(self, arrow: Style.ArrowStyle):
            for shape in self.composition:
                shape.style.arrow = arrow

        @property
        def shadow(self) -> float:
            """The shadow style of all ``Shapes`` in the composition."""
            return super().shadow

        @shadow.setter
        def shadow(self, shadow: float):
            for shape in self.composition:
                shape.style.shadow = shadow

    _shapes: Dict

    def __init__(self, shapes: Dict[str, Shape]):
        """shapes: list or dict of Shape objects."""
        super().__init__()
        self._shapes = shapes
        self._style = self.CompositionStyle(self)

    def __iter__(self):
        """Provides an ``iterable`` of the shapes in the composition."""
        return self._shapes.values().__iter__()

    def add(self, key: str, shape: Shape) -> "Composition":
        """Adds a shape to the composition.

        Args:
            key: the name of the shape to be added
            shape: the shape to be added.

        Returns:
            a copy of the composition with the requisite shape added
        """
        shapes = self._shapes.copy()
        shapes[key] = shape
        return Composition(shapes)

    def merge(self, comp: "Composition") -> "Composition":
        """Merge two compositions.

        Args:
            comp: the composition to merge with this one.

        Returns:
            a new composition made up of this one and the provided one.
        """
        return_composition = comp

        for key, shape in self._shapes.items():
            return_composition = return_composition.add(key, shape)
        return return_composition

    def __getitem__(self, key):
        """Gets the ``shape`` in the ``Composition`` with key ``key``."""
        return self._shapes[key]

    def _for_all_shapes(self, func: str, *args, **kwargs) -> "Composition":
        shapes = dict()
        for key, shape in self._shapes.items():
            shapes[key] = getattr(shape, func)(*args, **kwargs)
        return Composition(shapes)

    def rotate(self, angle: float, center: Point) -> "Composition":
        """Rotate the composition.

        Args:
            angle: The ``Angle`` in radians through which the shape should be rotated.
            center: The ``Point`` about which the rotation should be performed.

        Returns:
            A copy of the ``Composition`` rotated through ``angle`` about ``center``.
        """
        return self._for_all_shapes("rotate", angle, center)

    def translate(self, vec) -> "Composition":
        """Translate the composition.

        Args:
            vec: The vector through which the ``Composition`` should be translated.

        Returns:
            A copy of the ``Composition`` translated through ``vec``.
        """
        return self._for_all_shapes("translate", vec)

    def scale(self, factor) -> "Composition":
        """Scale the composition.

        Args:
            factor: The factor by which the ``Composition`` should be scaled.

        Returns:
            A copy of the ``Composition`` scaled by ``factor``.
        """
        return self._for_all_shapes("scale", factor)

    def apply(
        self, func: Callable[[Union[Shape, Text, "Composition"]], T]
    ) -> Dict[str, T]:
        """Applies a function to every member of a ``Composition``.

        Args:
            func: The function to apply.

        Returns:
            A dictionary of results.
        """
        ret_dict = {}
        for key, shape in self._shapes.items():
            ret_dict[key] = func(shape)
        return ret_dict



from typing import List

from pysketcher._curve import Curve
from pysketcher._point import Point
from pysketcher._style import Style


class Wall(Curve):
    """A representation of a wall.

    Args:
        points: a ``List`` of ``Point`` through which the wall should pass.
        thickness: the thickness of the wall.

    Examples:
        >>> model = ps.Wall(
        ...     [
        ...         ps.Point(1, 1),
        ...         ps.Point(2, 2),
        ...         ps.Point(3, 2.5),
        ...         ps.Point(4, 2),
        ...         ps.Point(5, 1),
        ...     ],
        ...     0.1,
        ... )
        >>> fig = ps.Figure(0, 6, 0, 3, backend=MatplotlibBackend)
        >>> fig.add(model)
        >>> fig.save("pysketcher/images/wall.png")

        .. figure:: images/wall.png
            :alt: An example of a Wall.
            :figclass: align-center

            An example ``Wall``.
    """

    _start: Point
    _end: Point
    _thickness: float

    def __init__(self, points: List[Point], thickness: float):
        self._start = points[0]
        self._end = points[-1]
        self._thickness = thickness

        def _displace(point: Point, point_before: Point, point_after: Point):
            # Displaces a point on our curve by thickness.
            # find a normal to the line between the point_before and the point_after
            # then displace by thickness from point in the direction of that normal
            return point + ((point_after - point_before).normal * self._thickness)

        # at the start, there isn't a point_before, so use the start point
        new_points: List[Point] = [_displace(points[0], points[0], points[1])]

        for i in range(1, len(points) - 1):
            new_points += [_displace(points[i], points[i - 1], points[i + 1])]

        # and at the end there isn't a point_after, so use the end point
        new_points += [_displace(points[-1], points[-2], points[-1])]

        points = points + new_points[-1::-1]
        points += [self._start]

        super().__init__(points)
        self.style.fill_pattern = Style.FillPattern.CROSS



import numpy as np

from pysketcher._angle import Angle
from pysketcher._arc import Arc
from pysketcher._point import Point


class Circle(Arc):
    """A representation of a 2D circle.

    Args:
        center: The center of the circle.
        radius: The radius of the circle.

    Examples:
        >>> circle = ps.Circle(ps.Point(1.5, 1.5), 1)
        >>> fig = ps.Figure(0, 3, 0, 3, backend=MatplotlibBackend)
        >>> fig.add(circle)
        >>> fig.save("pysketcher/images/circle.png")

    .. figure:: images/circle.png
        :alt: An example of Circle.
        :figclass: align-center

        An example of ``Circle``.
    """

    def __init__(self, center: Point, radius: float, resolution=180):
        super().__init__(center, radius, Angle(0), 2 * np.pi, resolution)



from typing import Callable, List, Tuple, Type, Union

from pysketcher._drawable import Drawable
from pysketcher.backend.backend import Backend


class Figure:
    """Provides the developer interface to the pysketcher backends.

    Provides the means to render models into either viewable, or
    savable (or both) images depending on which backend is chosen.

    Args:
        x_min: The minimum x-coordinate that will be rendered.
        x_max: The maximum x-coordinate that will be rendered.
        y_min: The minimum y-coordinate that will be rendered.
        y_max: The maximum y-coordinate that will be rendered.

    Examples:
        >>> circle = ps.Circle(ps.Point(1.5, 1.5), 1)
        >>> fig = ps.Figure(0, 3, 0, 3, backend=MatplotlibBackend)
        >>> fig.add(circle)
        >>> fig.save("pysketcher/images/figure.png")

        .. figure:: images/figure.png
            :alt: An example of the use of a Figure.
            :figclass: align-center

            An example of ``Figure``.
    """

    _shapes: List[Drawable]
    _backend: Backend

    def __init__(
        self,
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        backend: Type[Backend],
    ):
        self._backend = backend(x_min, x_max, y_min, y_max)
        self._shapes = []

    def add(self, shape: Drawable) -> None:
        """Adds a shape to the Figure.

        Once the shape is added, when :method:``save`` or :method:``show`` is
        called, the portion of the shape which is within the bounds of the figure
        will be included in the image.

        Args:
            shape: The shape which should be added to the figure.

        """
        self._shapes.append(shape)
        self._backend.add(shape)

    def show(self):
        """Shows an interactive view of the figure.

        If the backend supports it, then an interactive figure will be shown in a
        UI window.
        """
        self._backend.show()

    def save(self, filename: str) -> None:
        """Saves the rendered figure to a file.

        If the backed supports it, then an image will be saved to the location
        specified in ``filename``.

        Args:
            filename: The location to which the figure should be saved.
        """
        self._backend.save(filename)

    def erase(self):
        """Removes all the shapes from the figurnnnne."""
        self._backend.erase()

    def animate(
        self,
        func: Callable[[float], Drawable],
        interval: Union[Tuple[float, float], Tuple[float, float, float]],
    ):
        """Renders an animation from the provided function.

        If the backend supports it, then an animation will be rendered
        and stored internally by the backend.

        Args:
            func: This function takes a parameter and provides a `Drawable` for that
                parameter value.
            interval: the start and end values for the parameter and an optional
                increment. If the increment is not provided, the parameter will be
                incremented by 1.

        """
        self._backend.animate(func, interval)

    def save_animation(self, filename: str):
        """Saves a previously rendered animation to a file."""
        self._backend.save_animation(filename)



from pysketcher._angle import Angle
from pysketcher._point import Point
from pysketcher._shape import Shape
from pysketcher._style import TextStyle


class Text(Shape):
    """Place `text` on the drawing at the Point(x, y) `position`

    The `text` will be drawn in the given `direction`

    Args:
        text: The text to be displayed.
        position: Point, The position the text will be displayed at.
        direction: Point, The direction the text will flow to.

    Examples:
        >>> fig = ps.Figure(0.0, 4.0, 0.0, 4.0, MatplotlibBackend)
        >>> code = ps.Text("This is some left text!", Point(2, 1))
        >>> code.style.alignment = ps.TextStyle.Alignment.LEFT
        >>> code.style.line_color = ps.TextStyle.Color.BLUE
        >>> code.style.font_family = ps.TextStyle.FontFamily.SERIF
        >>> code1 = ps.Text("This is some right text!", Point(2, 2))
        >>> code1.style.alignment = ps.TextStyle.Alignment.RIGHT
        >>> code1.style.line_color = ps.TextStyle.Color.GREEN
        >>> code1.style.font_family = ps.TextStyle.FontFamily.SANS
        >>> code2 = ps.Text("This is some center text!", Point(2, 3))
        >>> code2.style.alignment = ps.TextStyle.Alignment.CENTER
        >>> code2.style.line_color = ps.TextStyle.Color.RED
        >>> code2.style.font_family = ps.TextStyle.FontFamily.MONO
        >>> fig.add(code)
        >>> fig.add(code1)
        >>> fig.add(code2)
        >>> fig.save("pysketcher/images/text.png")

        .. figure:: images/text.png
            :alt: An example of some text.
            :figclass: align-center

            An example of some ``Text``.
    """

    _style: TextStyle

    def __init__(
        self, text: str, position: Point, direction: Point = Point(1, 0)  # noqa: B008
    ):
        super().__init__()
        self._text: str = text
        self._position: Point = position
        self._direction: Point = direction
        self._style: TextStyle = TextStyle()

    def rotate(self, angle: Angle, center: Point) -> "Text":
        """Returns the text rotated through ``angle`` radians about ``centre``."""
        direction = self._direction.rotate(angle, center)
        position = self._position.rotate(angle, center)
        return Text(self._text, position, direction)

    def __str__(self):
        """Provides a string which describes the Text object."""
        return 'text "%s" at (%g,%g)' % (self._text, self._position.x, self._position.y)

    def __repr__(self):
        """Provides a string which describes the Text object."""
        return repr(str(self))

    @property
    def style(self) -> TextStyle:
        """Returns the style of the object so that the style can be altered."""
        return self._style

    @style.setter
    def style(self, text_style: TextStyle):
        self._style = text_style

    @property
    def position(self) -> Point:
        """The ``Point`` at which the text is located."""
        return self._position

    @property
    def direction(self) -> Point:
        """The direction in which the text flows."""
        return self._direction

    @property
    def text(self) -> str:
        """The text."""
        return self._text

    def set_alignment(self, alignment: TextStyle.Alignment) -> "Text":
        """Sets the alignment of the text.

        Args:
            alignment: The new alignment of the text.

        Returns:
            The original text object with the style modified to the new alignment.
        """
        self.style.alignment = alignment
        return self

    def translate(self, vec: Point) -> "Text":
        """Translates the text through ``vec``."""
        new_text = Text(self.text, self.position + vec, self.direction)
        new_text.style = self._style
        return new_text

    def scale(self, factor: float) -> "Text":
        """Scales the text by a factor of `factor`."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysketcher',
 'pysketcher._utils',
 'pysketcher.annotation',
 'pysketcher.backend',
 'pysketcher.backend.matplotlib',
 'pysketcher.composition',
 'pysketcher.dimension']

package_data = \
{'': ['*'], 'pysketcher': ['images/.gitignore']}

install_requires = \
['celluloid>=0.2.0,<0.3.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.19.5,<2.0.0',
 'scipy>=1.6.0,<2.0.0']

extras_require = \
{'docs': ['sphinx>=4.1.1,<5.0.0',
          'sphinx_rtd_theme>=0.5,<1.1',
          'sphinx-autodoc-typehints>=1.11.1,<2.0.0']}

entry_points = \
{'console_scripts': ['beam1 = examples.beam1:main']}

setup_kwargs = {
    'name': 'pysketcher',
    'version': '0.0.9',
    'description': 'Geometric Sketching Utility for Python',
    'long_description': '============\n PySketcher\n============\n\n.. image:: https://github.com/rvodden/pysketcher/workflows/Tests/badge.svg\n    :target: https://github.com/rvodden/pysketcher/actions?query=workflow%3ATests+branch%3Amaster\n\n.. image:: https://badgen.net/pypi/v/pysketcher?icon=pypi\n       :target: https://pypi.org/project/pysketcher/\n\n.. image:: https://api.codeclimate.com/v1/badges/eae2c2aa97080fbfed7e/maintainability\n    :target: https://codeclimate.com/github/rvodden/pysketcher/maintainability\n\n.. image:: https://codecov.io/gh/rvodden/pysketcher/branch/master/graph/badge.svg?token=AHCKOL75VY\n    :target: https://codecov.io/gh/rvodden/pysketcher\n\n.. image:: https://readthedocs.org/projects/pysketcher/badge/?version=latest&style=flat\n    :target: https://pysketcher.readthedocs.io/en/latest/\n\n.. image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n    :target: https://github.com/pre-commit/pre-commit\n\n.. image:: https://img.shields.io/badge/hypothesis-tested-brightgreen.svg\n    :target: https://hypothesis.readthedocs.io/\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n\n.. image:: https://badgen.net/github/dependabot/rvodden/pysketcher?icon=github\n    :target: https://github.com/rvodden/pysketcher\n\n**This is alpha software - the interface is likely to change with every release prior to 0.1.0.**\n\nPySketcher is a modern Python library designed to make creating geometric, mathematical and physical diagrams very\neasy.\n\nThis library is continues the legacy of Hans Petter Langtangen. Work done since he sadly passed in 2016 includes:\n\n1. The MatlibplotDraw object is no longer global and is no longer tightly coupled to the shape object. There is now a DrawingTool interface which this class implements.\n\n2. Code is organised into multiple files and published on pypi.\n\n3. Shapes are immutable. This means functions such as ``rotate`` return modified copies of the original shape, rather than altering the shape on which they are called.\n\n4. Angles are in radians not degrees.\n\n5. The Composition object is used more consistently. Previously objects such as Beam were direct children of Shape which led to code repetition.\n\n`Please see the documentation for more information <https://pysketcher.readthedocs.io/en/latest/index.html>`_.\n',
    'author': 'Richard Vodden',
    'author_email': 'richard@vodden.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rvodden/pysketcher',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from enum import auto, Enum, unique
from typing import Dict

import numpy as np

from pysketcher._arc import Arc
from pysketcher._line import Line
from pysketcher._point import Point
from pysketcher._shape import Shape
from pysketcher._style import Style
from pysketcher.annotation import ArcAnnotation
from pysketcher.composition import Composition


class AngularDimension(Composition):
    """Used to indicate the angle between two points about a given center.

    Args:
        start: the point from which the angle should be indicated
        end: the point to which the angle should be indicated
        center: the point about which the angle should be indicated

    returns: A composition.

    Examples:
        >>> arc1 = ps.Arc(ps.Point(0.25, 0.25), 1.0, ps.Angle(0.0), ps.Angle(np.pi / 2))
        >>> arc1.style.line_color = ps.Style.Color.BLUE
        >>> dim1 = ps.AngularDimension("$a$", arc1.start, arc1.end, arc1.center)
        >>>
        >>> arc2 = ps.Arc(
        ...     ps.Point(0.25, -0.25), 1.0, ps.Angle(0.0), ps.Angle(-np.pi / 2)
        ... )
        >>> arc2.style.line_color = ps.Style.Color.GREEN
        >>> dim2 = ps.AngularDimension("$b$", arc2.start, arc2.end, arc2.center)
        >>>
        >>> arc3 = ps.Arc(
        ...     ps.Point(-0.25, 0.25), 1.0, ps.Angle(np.pi / 2), ps.Angle(np.pi / 2)
        ... )
        >>> arc3.style.line_color = ps.Style.Color.RED
        >>> dim3 = ps.AngularDimension("$c$", arc3.start, arc3.end, arc3.center)
        >>>
        >>> fig = ps.Figure(-2.0, 2.0, -2.0, 2.0, backend=MatplotlibBackend)
        >>> fig.add(arc1)
        >>> fig.add(dim1)
        >>> fig.add(arc2)
        >>> fig.add(dim2)
        >>> fig.add(arc3)
        >>> fig.add(dim3)
        >>> fig.save("pysketcher/images/angular_dimension.png")
    """

    @unique
    class Orientation(Enum):
        """Specifies if the dimension should be drawn inside or outside the angle."""

        INTERNAL = auto()
        EXTERNAL = auto()

    _DEFAULT_OFFSET: float = 0.5
    _DEFAULT_MINOR_OFFSET: float = 0.1
    _DEFAULT_ORIENTATION: Orientation = Orientation.EXTERNAL
    _DEFAULT_EXTENSION_LINES: bool = True

    def __init__(self, text: str, start: Point, end: Point, center: Point):
        self._text = text
        self._start = start
        self._end = end
        self._center = center
        self._offset = self._DEFAULT_OFFSET
        self._minor_offset = self._DEFAULT_MINOR_OFFSET
        self._orientation = self._DEFAULT_ORIENTATION
        self._extension_lines = self._DEFAULT_EXTENSION_LINES
        super().__init__(self._generate_shapes())

    def _generate_shapes(self) -> Dict[str, Shape]:
        shapes = self._generate_extension_lines()

        # TODO: code a flag to indicate the outside angle
        self._start_angle = (self._start - self._center).angle
        self._end_angle = (self._end - self._center).angle
        if abs(self._end.angle - self._start.angle) > np.pi:
            self._start_angle, self._end_angle = self._end_angle, self._start_angle

        arc = Arc(
            self._center,
            abs(self._center - self._start) + self._offset,
            self._start_angle,
            (self._end_angle - self._start_angle),
        ).set_arrow(Style.ArrowStyle.DOUBLE)
        shapes["arrow"] = arc
        shapes["annotation"] = ArcAnnotation(self._text, arc)

        return shapes

    def _generate_extension_lines(self) -> Dict[str, Shape]:
        extension_lines = {}

        def extension_line_vector(p: Point, c: Point):
            if self._orientation == self.Orientation.EXTERNAL:
                vec = (p - c).unit_vector
            elif self._orientation == self.Orientation.INTERNAL:
                vec = (p - c).unit_vector * -1
            else:
                raise ValueError(f"Invalid value for Orientation: {self._orientation}")
            return vec

        if self._extension_lines:
            extension_line1_vector = extension_line_vector(self._start, self._center)
            extension_lines["extension_line_1"] = Line(
                self._start + extension_line1_vector * self._minor_offset,
                self._start
                + extension_line1_vector * (self._offset + self._minor_offset),
            )

            extension_line2_vector = extension_line_vector(self._end, self._center)
            extension_lines["extension_line_2"] = Line(
                self._end + extension_line2_vector * self._minor_offset,
                self._end
                + extension_line2_vector * (self._offset + self._minor_offset),
            )
        return extension_lines

    @property
    def start(self) -> Point:
        """The start of the dimension."""
        return self._start

    @property
    def end(self) -> Point:
        """The end of the dimension."""
        return self._end

    @property
    def center(self) -> Point:
        """The center of the dimension."""
        return self._center

    @property
    def extension_lines(self) -> bool:
        """If true, extension lines will be drawn."""
        return self._extension_lines

    @extension_lines.setter
    def extension_lines(self, extension_lines):
        self._extension_lines = extension_lines
        self._shapes = self._generate_shapes()



import numpy as np

from pysketcher._angle import Angle
from pysketcher._curve import Curve
from pysketcher._point import Point


class Arc(Curve):
    """A representation of a continuous connected subset of a circle.

    Args:
        center: The center of the Arc.
        radius: The radius of the Arc.
        start_angle: The angle from the +ve horizontal where the Arc should start.
        arc_angle: The angle (from the start_angle) where the Arc should end.
        resolution: The number of points in the Arc.

    Examples:
        >>> arc = ps.Arc(ps.Point(0.0, 0.0), 1.0, Angle(0.0), Angle(np.pi / 2))
        >>> fig = ps.Figure(-0.5, 1.5, -0.5, 1.5, backend=MatplotlibBackend)
        >>> fig.add(arc)
        >>> fig.save("pysketcher/images/arc.png")

    .. figure:: images/arc.png
        :alt: An example of an Arc.
        :figclass: align-center

        An example of an ``Arc``.
    """

    _center: Point
    _radius: float
    _start_angle: Angle
    _arc_angle: Angle
    _resolution: int

    def __init__(
        self,
        center: Point,
        radius: float,
        start_angle: Angle,
        arc_angle: Angle,
        resolution: int = 180,
    ):
        # Must record some parameters for __call__
        self._center = center
        self._radius = radius
        self._start_angle = Angle(start_angle)
        self._arc_angle = Angle(arc_angle)
        self._resolution = resolution

        if self._arc_angle == 0.0:
            # assume a full circle
            ts = np.linspace(0.0, 2.0 * np.pi, resolution + 1)
        else:
            ts = np.linspace(0.0, self._arc_angle, resolution + 1)

        points = [self(t) for t in ts]
        super().__init__(points)

    def __call__(self, theta: Angle) -> Point:
        """Provides a point on the arc ``theta`` of the way around.

        Args:
            theta: The angle from the ``start_angle`` from which the point should
                be taken.

        Returns:
            the point ``theta`` of the way around the arc.

        Raises:
            ValueError: if ``theta`` is beyond the bounds of the arc.
        """
        if 0.0 < self._arc_angle < theta or theta < self._arc_angle < 0.0:
            raise ValueError(
                f"Theta ({theta}) is outside the bounds "
                "of the arc (0.0 , {self._arc_angle})"
            )

        iota = Angle(self.start_angle + theta)
        ret_point = Point(
            self.center.x + self.radius * np.cos(iota),
            self.center.y + self.radius * np.sin(iota),
        )
        return ret_point

    @property
    def start_angle(self) -> Angle:
        """The angle from the +ve horizontal from where the arc starts."""
        return self._start_angle

    @property
    def arc_angle(self) -> Angle:
        """The angle from the ``start_angle`` to which the arc ends."""
        return self._arc_angle

    @property
    def end_angle(self) -> Angle:
        """The angle from the +ve horizontal to which the arc ends."""
        return self._start_angle + self._arc_angle

    @property
    def radius(self) -> float:
        """The radius of the arc."""
        return self._radius

    @property
    def center(self) -> Point:
        """The center of the arc."""
        return self._center

    @property
    def start(self) -> Point:
        """The point at which the arc starts."""
        return self(0.0)

    @property
    def end(self) -> Point:
        """The point at which the arc ends."""
        return self(self.arc_angle)

    @property
    def mid(self) -> Point:
        """The middle of the arc."""
        return self(self.arc_angle / 2)

    def translate(self, vec: Point) -> "Arc":
        """Translates the arc by the specified vector.

        Args:
            vec: The vector through which the arc should be translated.

        Returns:
            A copy of the arc which has been translated by ``vec``.
        """
        arc = Arc(
            self._center + vec,
            self._radius,
            self._start_angle,
            self._arc_angle,
            self._resolution,
        )
        arc.style = self.style
        return arc



from pysketcher._arc import Arc
from pysketcher._text import Text
from pysketcher.annotation._text_position import TextPosition


class ArcAnnotation(Text):
    """Annotates an arc with the provided text."""

    # TODO: write an example for ArcAnnotation
    _DEFAULT_OFFSET = 0.25

    def __init__(
        self, text: str, arc: Arc, text_position: TextPosition = TextPosition.MIDDLE
    ):
        self._arc = arc
        self._offset = self._DEFAULT_OFFSET
        radial = (arc.center - arc.mid).unit_vector
        text_pos = arc.mid - (radial * self._offset)
        super().__init__(text, text_pos)



"""Tool for creating sketches of physical and mathematical problems in Python code."""
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    pass

from pysketcher._angle import Angle
from pysketcher._arc import Arc
from pysketcher._arrow import Arrow, DoubleArrow
from pysketcher._axis import Axis
from pysketcher._circle import Circle
from pysketcher._cubic_bezier_curve import CubicBezier
from pysketcher._curve import Curve
from pysketcher._dashpot import Dashpot
from pysketcher._drawable import Drawable
from pysketcher._figure import Figure
from pysketcher._force import Force, Gravity
from pysketcher._line import Line
from pysketcher._moment import Moment
from pysketcher._point import Point
from pysketcher._rectangle import Rectangle
from pysketcher._shape import Shape
from pysketcher._simple_support import SimpleSupport
from pysketcher._sketchy_func import (
    SketchyFunc1,
    SketchyFunc2,
    SketchyFunc3,
    SketchyFunc4,
)
from pysketcher._spline import Spline
from pysketcher._spring import Spring
from pysketcher._style import Style, TextStyle
from pysketcher._text import Text
from pysketcher._triangle import Triangle
from pysketcher._uniform_load import UniformLoad
from pysketcher._velocity_profile import VelocityProfile
from pysketcher._wall import Wall
from pysketcher._wheel import Wheel
from pysketcher.annotation import ArcAnnotation, LineAnnotation, TextPosition
from pysketcher.composition import Composition
from pysketcher.dimension import AngularDimension, LinearDimension, RadialDimension

__all__ = [
    "AngularDimension",
    "Axis",
    "Angle",
    "Arc",
    "ArcAnnotation",
    "Arrow",
    "DoubleArrow",
    "Circle",
    "CubicBezier",
    "Composition",
    "Curve",
    "Dashpot",
    "Drawable",
    "Figure",
    "Force",
    "Gravity",
    "Line",
    "LineAnnotation",
    "LinearDimension",
    "Moment",
    "Point",
    "RadialDimension",
    "Rectangle",
    "Shape",
    "SimpleSupport",
    "SketchyFunc1",
    "SketchyFunc2",
    "SketchyFunc3",
    "SketchyFunc4",
    "Spline",
    "Spring",
    "Style",
    "TextStyle",
    "TextPosition",
    "Text",
    "Triangle",
    "UniformLoad",
    "VelocityProfile",
    "Wall",
    "Wheel",
]



from copy import copy
from typing import Tuple

from pysketcher._curve import Curve
from pysketcher._point import Point


class Line(Curve):
    """A representation of a line primitive.

    Args:
        start: The starting point of the line.
        end: The end point of the line.

    Example:
        >>> a = ps.Line(ps.Point(1.0, 2.0), ps.Point(4.0, 3.0))
        >>> b = a.rotate(np.pi / 2, ps.Point(1.0, 2.0))
        >>> fig = ps.Figure(0, 5, 0, 5, backend=MatplotlibBackend)
        >>> fig.add(a)
        >>> fig.add(b)
        >>> fig.save("pysketcher/images/line.png")

    .. figure:: images/line.png
        :alt: An example of Line.
        :figclass: align-center

        An example of ``Line``.
    """

    _start: Point
    _end: Point
    _a: float
    _b: float
    _c: float
    _d: float
    _vertical: bool
    _horizontal: bool

    def __init__(self, start: Point, end: Point):
        if start == end:
            raise ValueError("Cannot specify a line with two equal points.")
        self._start = start
        self._end = end
        self._horizontal = False
        self._vertical = False
        self._a = self._b = self._c = self._d = None
        super().__init__([self._start, self._end])
        self._compute_formulas()

    def _compute_formulas(self):
        # Define equations for line:
        # y = a*x + b,  x = c*y + d
        try:
            self._a = (self._end.y - self._start.y) / (self._end.x - self._start.x)
            self._b = self._start.y - self._a * self._start.x
        except ZeroDivisionError:
            self._vertical = True
            self._c = 0.0
            self._d = self._end.x
            return

        try:
            self._c = 1.0 / self._a
            self._d = self._b / self._a
        except ZeroDivisionError:
            self._horizontal = True

    @property
    def start(self) -> Point:
        """The starting point of the line."""
        return self._start

    @property
    def end(self) -> Point:
        """The end point of the line."""
        return self._end

    def __call__(self, x: float = None, y: float = None):
        """Given x, return y on the line, or given y, return x.

        Args:
            x: the value of x for which a value of y should be calculated
            y: the value of y for which a value of x should be calculated

        Returns:
            the appropriate value of either y or x.

        Raises:
            ValueError: If the line is horizontal and y is provided, or if
                the line is vertical and x is provided.
        """
        self._compute_formulas()
        if self._horizontal and y:
            raise ValueError("Value of x is not dependent on the value of y.")
        if self._vertical and x:
            raise ValueError("Value of y is not dependent on the value of x.")
        return self._a * x + self._b if x else self._c * y + self._d

    def interval(
        self, x_range: Tuple[float, float] = None, y_range: Tuple[float, float] = None
    ):
        """Returns a smaller portion of a line.

        Args:
            x_range: The range of x-coordinates which
                should be used to obtain the segment
            y_range: The range of y-coordinates which
                should be used to obtain the segment

        Returns:
            A line bounded to either ``x_range`` or ``y_range``.

        Raises:
            ValueError: If the line is vertical and ``x_range`` is provided, or if
                the line is horizontal and ``y_range`` is provided.
        """
        if x_range and y_range:
            raise ValueError("Cannot specify both x_range and y_range.")
        if x_range is not None:
            return Line(
                Point(x_range[0], self(x_range[0])), Point(x_range[1], self(x_range[1]))
            )
        elif y_range is not None:
            return Line(
                Point(self(y_range[0]), y_range[0]), Point(self(y_range[1]), y_range[1])
            )

    def rotate(self, angle: float, center: Point) -> "Line":
        """Rotate the line through an angle about a point.

        Args:
            angle: the angle in radians through which the rotation should occur.
            center: the point about which the rotation should occur.

        Returns:
            A copy of the line rotated through ``angle`` about ``center``.
        """
        start = self._start.rotate(angle, center)
        end = self._end.rotate(angle, center)
        line = Line(start, end)
        line.style = copy(self.style)
        return line



from enum import Enum
from typing import Any


class DocEnum(Enum):
    """Provides a handy way to document Enums."""

    def __new__(cls, value: Any, doc: str = None):
        """Creates an instance of a DocEnum.

        Args:
            value: The enumerate value as would be included in a normal enum.
            doc: A documentation string for the value.

        Returns:
            An instance of the DocEnum class.
        """
        self = object.__new__(cls)  # calling super().__new__(value) here would fail
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self



from pysketcher._angle import Angle
from pysketcher._arrow import Arrow
from pysketcher._point import Point
from pysketcher.annotation import LineAnnotation
from pysketcher.composition import Composition


class Axis(Composition):
    """A representation of a axis.

    Draw axis from start with `length` to the right
    (x axis). Place label at the end of the arrow tip.
    Then return `rotation_angle` (in degrees).
    The `label_spacing` denotes the space between the label
    and the arrow tip as a fraction of the length of the plot
    in x direction. A tuple can be given to adjust the position
    in both the x and y directions (with one parameter, the
    x position is adjusted).

    Args:
        start: The start of the ``Axis``.
        length: The length of the ``Axis``.
        label: A text label for the ``Axis``.
        rotation_angle: The ``Angle``
    """

    def __init__(
        self,
        start: Point,
        length: float,
        label: str,
        rotation_angle: Angle = Angle(0.0),  # noqa: B008
    ):
        arrow = Arrow(start, start + Point(length, 0)).rotate(rotation_angle, start)
        # should increase spacing for downward pointing axis
        label = LineAnnotation(label, arrow)

        super().__init__({"arrow": arrow, "label": label})



from pysketcher._line import Line
from pysketcher._point import Point
from pysketcher._style import Style


class Arrow(Line):
    """Draw an arrow as Line with arrow pointing towards end.

    Args:
        start: The start of the arrow.
        end: The end of the arrow.

    Examples:
        >>> up_arrow = ps.Arrow(ps.Point(0.0, 1.0), ps.Point(0.0, 3.0))
        >>> down_arrow = ps.Arrow(ps.Point(0.0, -1.0), ps.Point(0.0, -3.0))
        >>> left_arrow = ps.Arrow(ps.Point(-1.0, 0.0), ps.Point(-3.0, 0.0))
        >>> right_arrow = ps.Arrow(
        ...     ps.Point(1.0, 0.0),
        ...     ps.Point(3.0, 0.0),
        ... )
        >>> fig = ps.Figure(-4.0, 4.0, -4.0, 4.0, backend=MatplotlibBackend)
        >>> fig.add(up_arrow)
        >>> fig.add(down_arrow)
        >>> fig.add(left_arrow)
        >>> fig.add(right_arrow)
        >>> fig.save("pysketcher/images/arrow.png")

    .. figure:: images/arrow.png
        :alt: An example of Arrow.
        :figclass: align-center

        An example of ``Arrow``.
    """

    def __init__(self, start: Point, end: Point):
        super().__init__(start, end)
        self.style.arrow = Style.ArrowStyle.END


class DoubleArrow(Line):
    """Draw an arrow as Line with arrow pointing towards start and end.

    Args:
        start: The start of the double arrow.
        end: The end of the double arrow.

    Examples:
        >>> double_arrow = ps.DoubleArrow(
        ...     ps.Point(1.0, 1.0),
        ...     ps.Point(
        ...         3.0,
        ...         1.0,
        ...     ),
        ... )
        >>> fig = ps.Figure(0.0, 4.0, 0.0, 2.0, backend=MatplotlibBackend)
        >>> fig.add(double_arrow)
        >>> fig.save("pysketcher/images/double_arrow.png")

    .. figure:: images/double_arrow.png
        :alt: An example of DoubleArrow.
        :figclass: align-center

        An example of ``DoubleArrow``.
    """

    def __init__(self, start: Point, end: Point):
        super().__init__(start, end)
        self.style.arrow = Style.ArrowStyle.DOUBLE
